import pathlib
import shutil

import mlflow.exceptions
import pytest
import yaml

import mantik.config.environment as environment
import mantik.config.exceptions as exceptions
import mantik.config.validate as project_config


@pytest.fixture(autouse=True)
def set_required_env_vars(env_vars_set, required_config_env_vars) -> None:
    with env_vars_set(required_config_env_vars):
        yield


class TestConfigValidator:
    @pytest.mark.parametrize(
        ("mlproject_path", "config_path", "expected"),
        [
            # Test case: config path is not absolute
            (
                "/absolute/path",
                "relative/path",
                pathlib.Path("/absolute/path/relative/path"),
            ),
            # Test case: both paths are absolute
            (
                "/absolute/path",
                "/absolute/path/config",
                pathlib.Path("/absolute/path/config"),
            ),
        ],
    )
    def test_config_absolute_path(self, mlproject_path, config_path, expected):
        config = project_config.ProjectValidator(
            mlproject_path, config_path, {}, "main"
        )
        assert config.config_absolute_path == expected

    @pytest.mark.parametrize(
        ("mlproject_path", "config_path", "expected"),
        [
            # Test case: config path is not absolute
            ("/absolute/path", "relative/path", pathlib.Path("relative/path")),
            # Test case: both paths are absolute
            ("/absolute/path", "/absolute/path/config", pathlib.Path("config")),
        ],
    )
    def test_config_relative_path(self, mlproject_path, config_path, expected):
        config = project_config.ProjectValidator(
            mlproject_path, config_path, {}, "main"
        )
        assert config.config_relative_path == expected

    @pytest.mark.parametrize(
        ("fake_mlproject_path", "config_path", "expected"),
        [
            # Test case: MLproject path is not absolute
            (
                "absolute/path",
                "relative/path",
                "ML project path must be an absolute path, "
                "but 'absolute/path' was given.",
            ),
            # Test case: MLproject path does not exist
            (
                "/absolute/path",
                "relative/path",
                "ML project path not found at '/absolute/path', "
                "check that the given path is correct.",
            ),
            # Test case: config path does not exist
            (
                "ml_project_resource",
                "false/path/to/config.yml",
                "Config not found at "
                "'ml_project_resource/false/path/to/config.yml', "
                "check that the given path is correct.",
            ),
            # Test case: config path is from another project
            (
                "ml_project_resource",
                "invalid_config_type",
                "Config file is not in the ML project directory, "
                "check that the given paths "
                "are correct:\nConfig file: "
                "'invalid_config_type'\nML project directory: "
                "'ml_project_resource'",
            ),
            # Test case: config with errors
            (
                "ml_project_resource",
                "config_with_errors",
                "Either UnicoreApiUrl or Firecrest details must be provided",
            ),
            # Test case: config without Environment section validation
            # successful
            (
                "ml_project_resource",
                "compute_backend_config_yaml",
                None,
            ),
            # Test case: config with Environment section validation successful
            (
                "ml_project_resource",
                "compute_backend_config_json",
                None,
            ),
        ],
    )
    def test_validate(
        self,
        expect_raise_if_exception,
        fake_mlproject_path,
        config_path,
        expected,
        mlproject_path,
        invalid_config_type,
        config_with_errors,
        compute_backend_config_yaml,
        compute_backend_config_json,
    ):
        expected_error = (
            exceptions.ConfigValidationError() if expected is not None else None
        )

        with expect_raise_if_exception(expected_error) as e:
            final_mlproject_path = fake_mlproject_path
            if fake_mlproject_path == "ml_project_resource":
                final_mlproject_path = mlproject_path

            if config_path == "invalid_config_type":
                config_path = invalid_config_type
            elif config_path == "config_with_errors":
                config_path = config_with_errors
            elif config_path == "compute_backend_config_yaml":
                config_path = compute_backend_config_yaml
            elif config_path == "compute_backend_config_json":
                config_path = compute_backend_config_json

            mlproject_config = project_config.ProjectValidator(
                final_mlproject_path, config_path, {}, "main"
            )
            mlproject_config.validate()

        if expected is not None:
            expected = (
                expected.replace(
                    "ml_project_resource", mlproject_path.as_posix()
                )
                .replace("invalid_config_type", invalid_config_type.as_posix())
                .replace("config_with_errors", config_with_errors.as_posix())
            )

            result = str(e.value)

            assert result == expected

    @pytest.mark.parametrize(
        ("execution_dict", "expected", "error_message"),
        [
            # Test case: local apptainer path does not exist
            (
                {
                    "Apptainer": {
                        "Path": "non-existent-image",
                        "Type": "local",
                    }
                },
                exceptions.ConfigValidationError(),
                "The path '"
                "mlproject_path/non-existent-image' "
                "given as execution environment was not found, "
                "check that the given path is correct. "
                "The path must be relative to the ML project path.",
            ),
            # Test case: image exists
            (
                {"Apptainer": {"Path": "mantik-test.sif", "Type": "local"}},
                None,
                None,
            ),
            # Test case: image is remote and path absolute
            (
                {
                    "Apptainer": {
                        "Path": "/absolute/remote/image",
                        "Type": "remote",
                    }
                },
                None,
                None,
            ),
            # Test case: image is remote and path relative
            (
                {
                    "Apptainer": {
                        "Path": "relative/remote/image",
                        "Type": "remote",
                    }
                },
                exceptions.ConfigValidationError(),
                "If image type 'remote' is given for an Apptainer image, "
                "the given path must be absolute. "
                "The given path is: 'relative/remote/image'.",
            ),
            # Test case: python venv has relative path
            (
                {
                    "Python": {
                        "Path": "relative/remote/image",
                    }
                },
                exceptions.ConfigValidationError(),
                "The given path to the Python Venv must be absolute. "
                "The given path is: 'relative/remote/image'.",
            ),
        ],
    )
    def test_validate_execution_path(
        self,
        expect_raise_if_exception,
        mlproject_path,
        execution_dict,
        expected,
        error_message,
    ):
        with expect_raise_if_exception(expected) as e:
            config = project_config.ProjectValidator(
                mlproject_path, "config_path", {}, "main"
            )
            execution = environment._get_execution_environment(execution_dict)

            result = config._validate_execution_path(execution)
            assert result is expected

        if error_message is not None:
            expected_error = error_message.replace(
                "mlproject_path", mlproject_path.as_posix()
            )
            result_error = str(e.value)
            assert result_error == expected_error

    @pytest.mark.parametrize(
        ("entry_point", "parameters", "expected"),
        [
            (
                "this-entry-point-does-not-exist",
                {},
                mlflow.exceptions.ExecutionException(
                    "Could not find this-entry-point-does-not-exist "
                    "among entry points ['main'] "
                    "or interpret this-entry-point-does-not-exist "
                    "as a runnable script. "
                    "Supported script file extensions: ['.py', '.sh']"
                ),
            ),
            ("main", {}, None),
        ],
    )
    def test_validate_ml_project_file(
        self,
        expect_raise_if_exception,
        mlproject_path,
        entry_point,
        parameters,
        expected,
    ):
        with expect_raise_if_exception(expected) as e:
            config = project_config.ProjectValidator(
                mlproject_path,
                "compute-backend-config.yaml",
                parameters,
                entry_point,
            )
            config.validate()
            if expected is not None:
                assert str(e.value) == expected.message

    def test_validate_config_not_in_exclude(
        self,
        mlproject_path,
        tmp_path,
        expect_raise_if_exception,
    ):
        config_file_name = "compute-backend-config.yaml"
        path = _copy_mlproject_and_add_exclude_to_config(
            tmp_path=tmp_path,
            mlproject_path=mlproject_path,
            name=config_file_name,
        )

        with expect_raise_if_exception(exceptions.ConfigValidationError()) as e:
            config = project_config.ProjectValidator(
                path, config_file_name, {}, "main"
            )
            config.validate()

        assert f"Config file '{config_file_name}' cannot be excluded" in str(
            e.value
        )


def _copy_mlproject_and_add_exclude_to_config(
    tmp_path: pathlib.Path, mlproject_path: pathlib.Path, name: str
) -> pathlib.Path:
    target = tmp_path / "mlproject"
    shutil.copytree(mlproject_path, target)

    with open(mlproject_path / name) as f:
        config = yaml.safe_load(f.read())
        config["Exclude"] = [name]

    with open(target / name, "w") as f:
        yaml.safe_dump(config, stream=f, indent=4)

    return target



import uuid

import click


DATA_REPOSITORY_NAME = click.option(
    "--name",
    required=False,
    default=None,
    type=str,
    help="Data repository name, displayed in Mantik. "
    "When not passed the TARGET is passed as name.",
)

# TODO this will probably in the future need to be refactored
#      to look more similar to the similar options in mantik/cli/_options.py
#      to have more cohesion in the api

DATA_REPOSITORY_ID = click.option(
    "--id",
    required=False,
    default=None,
    type=uuid.UUID,
    help="Data repository ID.",
)

DATA_REPOSITORY_DESCRIPTION = click.option(
    "--description",
    required=False,
    default=None,
    type=str,
    help="Data repository description, displayed in Mantik.",
)



"""Integration tests for UNICORE.

**WARNING:** The UNICORE container takes up to 15-30 seconds to process a job,
hence be careful with the number of test cases used.


"""
import pathlib
import uuid

import mlflow.entities as entities
import mlflow.projects as projects
import pytest

import mantik.config.core as core
import mantik.utils as utils
import mantik.utils.credentials as _credentials
import mantik_compute_backend.backend as backend

FILE_PATH = pathlib.Path(__file__).parent


@pytest.fixture()
def required_unicore_env_vars(unicore_credentials) -> dict:
    return {
        _credentials.UNICORE_USERNAME_ENV_VAR: unicore_credentials["user"],
        _credentials.UNICORE_PASSWORD_ENV_VAR: unicore_credentials["password"],
        core.COMPUTE_BUDGET_ACCOUNT_ENV_VAR: "test-project",
        utils.mlflow.ACTIVE_RUN_ID_ENV_VAR: str(uuid.uuid4().hex),
    }


RESOURCES = {
    "Queue": "batch",
    "Runtime": "1h",
    "Nodes": 1,
    "TotalCPUs": 1,
    "CPUsPerNode": 1,
    "MemoryPerNode": "1M",
}

EXPECTED_RESOURCE_MESSAGES = (
    "Queue=batch",
    "Runtime=3600",
    "Nodes=1",
    "TotalCPUs=1",
    "CPUsPerNode=1",
    "MemoryPerNode=1048576",
    "Project=test-project",
)


class TestUnicoreBackend:
    @pytest.mark.parametrize(
        (
            "entry_point",
            "params",
            "backend_config",
            "expected_job_status",
            "expected_log_messages",
        ),
        [
            # Test case: Apptainer with 1 env var
            #
            # * Test case will fail since no Apptainer image present in UNICORE
            #   container, thus the post command on login node won't be executed
            (
                "main",
                {"print": "test"},
                {
                    "Environment": {
                        "Variables": {"TEST_VAR": "test_value"},
                        "Apptainer": {
                            "Path": "/does/not/exist.sif",
                            "Type": "remote",
                        },
                        "PreRunCommandOnLoginNode": [
                            "echo precommand login node",
                            "echo again",
                        ],
                        "PostRunCommandOnLoginNode": [
                            "echo postcommand login node"
                        ],
                        "PreRunCommandOnComputeNode": [
                            "echo precommand compute node"
                        ],
                        "PostRunCommandOnComputeNode": [
                            "echo postcommand compute node"
                        ],
                    },
                    "Resources": RESOURCES,
                },
                entities.RunStatus.FAILED,
                [
                    *EXPECTED_RESOURCE_MESSAGES,
                    "MLFLOW_RUN_ID=",
                    "; export MLFLOW_RUN_ID",
                    'TEST_VAR="test_value"; export TEST_VAR',
                    'source $(dirname "$(realpath "$0")")/mantik.sh',
                ],
            ),
            # Test case: Python with 1 env var
            #
            # * Test case will fail since no venv present in UNICORE container,
            #   thus the post command on login node won't be executed
            (
                "main",
                {"print": "test"},
                {
                    "Environment": {
                        "Variables": {"TEST_VAR": "test_value"},
                        "Python": {
                            "Path": "/does/not/exist",
                        },
                        "PreRunCommandOnLoginNode": [
                            "echo precommand login node",
                            "echo again",
                        ],
                        "PostRunCommandOnLoginNode": [
                            "echo postcommand login node"
                        ],
                        "PreRunCommandOnComputeNode": [
                            "echo precommand compute node"
                        ],
                        "PostRunCommandOnComputeNode": [
                            "echo postcommand compute node"
                        ],
                    },
                    "Resources": RESOURCES,
                },
                entities.RunStatus.FAILED,
                [
                    *EXPECTED_RESOURCE_MESSAGES,
                    "MLFLOW_RUN_ID=",
                    "; export MLFLOW_RUN_ID",
                    'TEST_VAR="test_value"; export TEST_VAR',
                    "echo precommand login node",
                    "echo again",
                    'source $(dirname "$(realpath "$0")")/mantik.sh',
                ],
            ),
            # Test case: No execution environment and echo entry point
            #
            # * Should succeed
            # * Pre/postcommand will be executed on login node, hence no modules
            #   loaded here. Modules would make the pre command fail (no modules
            #   in UNICORE container), and the main application will not be
            #   executed, as well as the post command.
            (
                "echo",
                {"output": "test"},
                {
                    "Environment": {
                        "Variables": {"TEST_VAR": "test_value"},
                        "PreRunCommandOnLoginNode": [
                            "echo precommand login node",
                            "echo again",
                        ],
                        "PostRunCommandOnLoginNode": [
                            "echo postcommand login node"
                        ],
                    },
                    "Resources": RESOURCES,
                },
                entities.RunStatus.FINISHED,
                [
                    "Launched pre command <echo precommand login node && "
                    "echo again>\n",
                    "Executing command: echo precommand login node && "
                    "echo again",
                    "Execution on login node",
                    *EXPECTED_RESOURCE_MESSAGES,
                    'TEST_VAR="test_value"; export TEST_VAR',
                    "MLFLOW_RUN_ID=",
                    "; export MLFLOW_RUN_ID",
                    'source $(dirname "$(realpath "$0")")/mantik.sh',
                ],
            ),
        ],
    )
    def test_run(
        self,
        tmp_path,
        env_vars_set,
        unicore_api_url,
        required_unicore_env_vars,
        example_project_path,
        entry_point,
        params,
        backend_config,
        expected_job_status,
        expected_log_messages,
    ):
        backend_config = {
            **backend_config,
            "UnicoreApiUrl": unicore_api_url,
            # Following values are set by MLflow before running a project.
            projects.PROJECT_STORAGE_DIR: example_project_path.as_posix(),
        }

        expected_log_messages.extend(
            [
                # MLflow Tracking URI should be set by compute backend
                f'MLFLOW_TRACKING_URI="file://{tmp_path.as_posix()}/mlruns"; export MLFLOW_TRACKING_URI',  # noqa: E501
            ]
        )

        # Following env vars must be set for the config.
        tracking_uri = tmp_path / "mlruns"
        with env_vars_set(
            {
                **required_unicore_env_vars,
                # Point MLFLOW_TRACKING_URI to a temporary directory
                utils.mlflow.TRACKING_URI_ENV_VAR: tracking_uri.as_uri(),  # noqa: E501
            }
        ):
            result = backend.Backend().run(
                project_uri=example_project_path.as_posix(),
                entry_point=entry_point,
                params=params,
                version=None,
                backend_config=backend_config,
                tracking_uri=None,
                experiment_id=None,
            )
            # Wait for job to fail or succeed
            result.wait()
            # For the structure of the UNICORE job properties responce, see
            # `tests/resources/unicore-responses/job-property-response.json`.
            result_log_messages = "\n".join(result._job._job.properties["log"])
        for exp in expected_log_messages:
            assert exp in result_log_messages, f"Message {exp!r} not in logs"
        assert result.get_status() == expected_job_status



import datetime
import pathlib

import freezegun
import pytest

import mantik.authentication.tokens as _tokens


@pytest.fixture()
def tokens():
    return _tokens.Tokens(
        access_token="test-access-token",
        refresh_token="test-refresh-token",
        expires_at=datetime.datetime(
            2000, 1, 1, 1, 30, tzinfo=datetime.timezone.utc
        ),
    )


class TestTokens:
    @pytest.mark.parametrize(
        ("current_date", "expected"),
        [
            ("2000-01-01T01:00", False),
            ("2000-01-01T02:00", True),
        ],
    )
    def test_has_expired(self, tokens, current_date, expected):
        with freezegun.freeze_time(current_date):
            result = tokens.has_expired

            assert result == expected

    @pytest.mark.parametrize(
        "path",
        ["test-tokens.json", "subfolder/test-tokens.json"],
    )
    def test_write_to_and_from_file(self, tmp_path, tokens, path):
        file = tmp_path / path
        as_path = pathlib.Path(file)
        tokens.write_to_file(as_path)

        result = tokens.from_file(as_path)

        assert result == tokens



import pathlib

import mlflow.entities as entities
import pytest

import mantik.testing as testing
import mantik.testing.pyunicore as test_unicore
import mantik_compute_backend.firecrest._properties as _firecrest_properties
import mantik_compute_backend.firecrest.job as _job
import mantik_compute_backend.submitted_run as _submitted_run
import mantik_compute_backend.unicore._properties as _unicore_properties

FILE_DIR = pathlib.Path(__file__).parent


class TestSubmittedRunFirecrest:
    @pytest.mark.parametrize(
        "status", [e.value for e in _firecrest_properties.Status]
    )
    def test_get_status(self, status: str):
        job = _job.Job(
            client=testing.firecrest.FakeClient(job_status=status),
            machine="test-machine",
            job_id="test-job-id",
            job_dir=pathlib.Path("test-rundir"),
        )
        submitted_run = _submitted_run.SubmittedRun(
            run_id="test-run-id", job=job
        )
        submitted_run.get_status()

    def test_wait(self):
        job = _job.Job(
            client=testing.firecrest.FakeClient(
                job_status=_firecrest_properties.Status.COMPLETED.value
            ),
            machine="test-machine",
            job_id="test-job-id",
            job_dir=pathlib.Path("test-rundir"),
        )
        submitted_run = _submitted_run.SubmittedRun(
            run_id="test-run-id", job=job
        )
        submitted_run.wait()
        assert submitted_run.get_status() == entities.RunStatus.FINISHED


class TestSubmittedRunUnicore:
    def test_run_id(self):
        run = test_unicore._create_run()
        expected = "test-job"

        result = run.run_id

        assert result == expected

    @pytest.mark.parametrize(
        ("will_be_successful", "expected"),
        [
            (
                True,
                True,
            ),
            (
                False,
                False,
            ),
        ],
    )
    def test_wait(self, will_be_successful, expected):
        run = test_unicore._create_run(
            will_be_successful=will_be_successful,
        )
        result = run.wait()

        assert result == expected

    @pytest.mark.parametrize(
        ("status", "expected"),
        [
            (
                _unicore_properties.Status.STAGING_IN,
                entities.RunStatus.SCHEDULED,
            ),
            (
                _unicore_properties.Status.READY,
                entities.RunStatus.SCHEDULED,
            ),
            (
                _unicore_properties.Status.QUEUED,
                entities.RunStatus.SCHEDULED,
            ),
            (
                _unicore_properties.Status.RUNNING,
                entities.RunStatus.RUNNING,
            ),
            (
                _unicore_properties.Status.STAGING_OUT,
                entities.RunStatus.RUNNING,
            ),
            (
                _unicore_properties.Status.SUCCESSFUL,
                entities.RunStatus.FINISHED,
            ),
            (
                _unicore_properties.Status.FAILED,
                entities.RunStatus.FAILED,
            ),
            (
                _unicore_properties.Status.UNKNOWN,
                entities.RunStatus.RUNNING,
            ),
        ],
    )
    def test_get_status(self, status, expected):
        run = test_unicore._create_run(status=status)
        result = run.get_status()

        assert result == expected

    def test_cancel(self):
        run = test_unicore._create_run()

        run.cancel()



import logging
import os
import unittest.mock
import uuid

import click.testing
import pytest

import mantik.cli.main as main
import mantik.cli.models.models

GROUP_NAME = mantik.cli.models.models.GROUP_NAME

SAMPLE_UUID = uuid.uuid4()


@pytest.fixture
def sample_uuid_str() -> str:
    return str(SAMPLE_UUID)


@pytest.fixture
def mock_project_id_env_var(sample_uuid_str):
    os.environ[mantik.utils.env_vars.PROJECT_ID_ENV_VAR] = sample_uuid_str
    yield
    os.unsetenv(mantik.utils.env_vars.PROJECT_ID_ENV_VAR)


@pytest.fixture
def cli_test_runner():
    yield click.testing.CliRunner()


def test_get_all_project_models(
    mock_authentication, cli_test_runner, sample_uuid_str
):
    with unittest.mock.patch(
        "mantik.utils.mantik_api.models.get_all"
    ) as mocked_function:
        result = cli_test_runner.invoke(
            main.cli,
            [GROUP_NAME, "list", "--project-id", sample_uuid_str],
        )
        assert result.exit_code == 0
        mocked_function.assert_called()


def test_get_model(mock_authentication, cli_test_runner, sample_uuid_str):
    with unittest.mock.patch(
        "mantik.utils.mantik_api.models.get_one"
    ) as mocked_function:
        result = cli_test_runner.invoke(
            main.cli,
            [
                GROUP_NAME,
                "get-one",
                "--model-id",
                sample_uuid_str,
                "--project-id",
                sample_uuid_str,
            ],
        )
        assert result.exit_code == 0
        mocked_function.assert_called()


def test_delete_model(
    mock_authentication,
    mock_project_id_env_var,
    cli_test_runner,
    sample_uuid_str,
):
    with unittest.mock.patch(
        "mantik.utils.mantik_api.models.delete"
    ) as mocked_function:
        result = cli_test_runner.invoke(
            main.cli,
            [GROUP_NAME, "delete", "--model-id", sample_uuid_str],
        )
        assert result.exit_code == 0
        mocked_function.assert_called()


def test_create_model_entry(
    mock_authentication, cli_test_runner, sample_uuid_str
):
    with unittest.mock.patch(
        "mantik.utils.mantik_api.models.add"
    ) as mocked_function:
        result = cli_test_runner.invoke(
            main.cli,
            [
                GROUP_NAME,
                "add",
                "--project-id",
                sample_uuid_str,
                "--uri",
                "URI",
                "--location",
                "something://somewhere.com",
                "--name",
                "new-name",
            ],
        )
        assert result.exit_code == 0
        mocked_function.assert_called()


def test_update_model_entry(
    mock_authentication, cli_test_runner, sample_uuid_str
):
    with unittest.mock.patch(
        "mantik.utils.mantik_api.models.update"
    ) as mocked_function:
        result = cli_test_runner.invoke(
            main.cli,
            [
                GROUP_NAME,
                "update",
                "--project-id",
                sample_uuid_str,
                "--model-id",
                sample_uuid_str,
                "--uri",
                "URI",
                "--location",
                "something://somewhere.com",
                "--name",
                "new-name",
            ],
        )
        assert result.exit_code == 0
        mocked_function.assert_called()


@pytest.mark.parametrize(
    "mlflow_params,expected",
    [
        ('{"hello":"world"}', {"hello": "world"}),
        (None, None),
    ],
)
def test_mlflow_parameters_decoder(mlflow_params, expected):
    assert (
        mantik.cli.models.models.mlflow_parameters_decoder(
            None, None, mlflow_params
        )
        == expected
    )


def test_mlflow_parameters_decoder_raises_error_on_wrong_format():
    with pytest.raises(TypeError):
        mantik.cli.models.models.mlflow_parameters_decoder(
            None, None, "i am a broken json string"
        )


@pytest.fixture
def docker_client():
    class Client:
        class Images:
            def load(self, data):
                pass

        images = Images()

    return Client()


@pytest.fixture
def mock_get_image_url():
    yield unittest.mock.patch(
        "mantik.utils.mantik_api.models.get_image_url", return_value="test-url"
    )


@pytest.fixture
def mock_get_url():
    yield unittest.mock.patch("requests.get", return_value=[b"1", b"2", b"3"])


@pytest.fixture
def mock_docker_client(docker_client):
    yield unittest.mock.patch("docker.from_env", return_value=docker_client)


def test_download_model(
    mock_authentication,
    cli_test_runner,
    sample_uuid_str,
    tmpdir,
    mock_get_image_url,
    mock_get_url,
    caplog,
):
    caplog.set_level(logging.INFO)
    with mock_get_image_url as get_image_url, mock_get_url as get_url:
        result = cli_test_runner.invoke(
            main.cli,
            [
                GROUP_NAME,
                "download",
                "--project-id",
                sample_uuid_str,
                "--model-id",
                sample_uuid_str,
                "--target-dir",
                tmpdir,
            ],
        )
        assert result.exit_code == 0
        get_image_url.assert_called()
        get_url.assert_called()

        assert "Downloading image" in caplog.text
        assert "Image saved as zipped tarball at" in caplog.text
        assert (
            "In order to load the image into docker, "
            "you can run `docker load <"
        ) in caplog.text


def test_download_model_load(
    mock_authentication,
    cli_test_runner,
    sample_uuid_str,
    tmpdir,
    mock_get_image_url,
    mock_get_url,
    mock_docker_client,
    caplog,
):
    caplog.set_level(logging.INFO)
    with mock_get_image_url as get_image_url, mock_get_url as request, mock_docker_client as load:  # noqa E501
        result = cli_test_runner.invoke(
            main.cli,
            [
                GROUP_NAME,
                "download",
                "--project-id",
                sample_uuid_str,
                "--model-id",
                sample_uuid_str,
                "--target-dir",
                tmpdir,
                "--load",
            ],
        )
        assert result.exit_code == 0, result.output
        get_image_url.assert_called()
        request.assert_called()
        load.assert_called()

        assert "Downloading image" in caplog.text
        assert "Image saved as zipped tarball at" in caplog.text
        assert (
            "Unzipping and loading into docker. This can take a few minutes..."
            in caplog.text
        )
        assert (
            "Run 'docker images', an image named test-url should be present."
        ) in caplog.text


def test_download_model_apptainer(
    mock_authentication,
    cli_test_runner,
    sample_uuid_str,
    tmpdir,
    docker_client,
    caplog,
):
    result = cli_test_runner.invoke(
        main.cli,
        [
            GROUP_NAME,
            "download",
            "--project-id",
            sample_uuid_str,
            "--model-id",
            sample_uuid_str,
            "--target-dir",
            tmpdir,
            "--image-type",
            "apptainer",
        ],
    )
    assert result.exit_code == 1
    assert ("We are sorry, " "apptainer is not supported yet!\n") in caplog.text


def test_start_build(mock_authentication, cli_test_runner, sample_uuid_str):
    with unittest.mock.patch(
        "mantik.utils.mantik_api.models.start_build"
    ) as mocked_function:
        result = cli_test_runner.invoke(
            main.cli,
            [
                GROUP_NAME,
                "build",
                "--model-id",
                sample_uuid_str,
                "--project-id",
                sample_uuid_str,
            ],
        )
        assert result.exit_code == 0
        mocked_function.assert_called()


def test_get_build_status(
    mock_authentication, cli_test_runner, sample_uuid_str
):
    with unittest.mock.patch(
        "mantik.utils.mantik_api.models.get_one"
    ) as mocked_function:
        result = cli_test_runner.invoke(
            main.cli,
            [
                GROUP_NAME,
                "build-status",
                "--model-id",
                sample_uuid_str,
                "--project-id",
                sample_uuid_str,
            ],
        )
        assert result.exit_code == 0
        mocked_function.assert_called()
        assert ("Build Status") in result.output



import uuid

import pytest

import mantik.utils.mantik_api.connection as connection
import mantik.utils.mantik_api.user as user


@pytest.fixture(scope="session")
def fake_connection_id():
    return uuid.uuid4()


@pytest.fixture(scope="session")
def fake_user_id():
    return uuid.uuid4()


@pytest.fixture()
def mock_get_user_id_from_token(mocker, fake_user_id):
    with mocker.patch(
        "mantik.remote_file_service.s3_file_service._get_sub_from_token",
        return_value=str(fake_user_id),
    ):
        yield


@pytest.fixture()
def mock_login(mocker):
    with mocker.patch(
        "mantik.authentication.auth.get_valid_access_token", return_value="XXX"
    ):
        yield


@pytest.fixture()
def fake_connection(fake_connection_id, fake_user_id):
    return connection.Connection(
        connection_id=fake_connection_id,
        user=user.User(
            user_id=uuid.UUID("3fa85f64-5717-4562-b3fc-2c963f66afa6"),
            name="test-name",
        ),
        connection_name="s3 connection",
        connection_provider="ambrosys",
        auth_method="username/password",
        login_name="api-name",
        password="api-password",
    )


@pytest.fixture()
def mock_get_connection(mocker, fake_connection):
    with mocker.patch(
        "mantik.utils.mantik_api.connection.get",
        return_value=fake_connection,
    ):
        yield


@pytest.fixture()
def mock_api_creds_retrieval(
    mock_login, mock_get_connection, mock_get_user_id_from_token
):
    yield



import pathlib
import typing as t
import uuid

import click

import mantik.cli._options as _options
import mantik.cli.main as main
import mantik.cli.utils as cli_utils
import mantik.data_repository.data_repository as data_repository
import mantik.utils

GROUP_NAME = "data"


@main.cli.group(GROUP_NAME)
def cli() -> None:
    """Interaction with data through the mantik api."""


@cli.command("download")
@_options.PROJECT_ID
@_options.DATA_REPOSITORY_ID
@click.option(
    "--commit",
    required=False,
    type=str,
    help="Commit hash to checkout. Takes precedence over branch.",
    envvar=mantik.utils.env_vars.DATA_REPOSITORY_COMMIT_ENV_VAR,
    default=None,
)
@click.option(
    "--branch",
    required=False,
    type=str,
    help="Branch to checkout. Defaults main branch.",
    envvar=mantik.utils.env_vars.DATA_REPOSITORY_BRANCH_ENV_VAR,
    default=None,
)
@_options.get_target_dir_option_required(
    help_option="Path to directory where the data will be stored.",
)
def download_data_repository(
    project_id: uuid.UUID,
    data_repository_id: uuid.UUID,
    commit: t.Optional[str],
    branch: t.Optional[str],
    target_dir: str,
) -> None:
    """Download data repository."""
    output = data_repository.download_data_repository(
        project_id=project_id,
        data_repository_id=data_repository_id,
        checkout=commit or branch,
        target_dir=pathlib.Path(target_dir),
        token=cli_utils.access_token_from_env_vars(),
    )
    click.echo(output)



import contextlib
import datetime
import logging
import os
import pathlib
import typing as t
import uuid

import pytest
import requests_mock

import mantik.authentication.auth as track
import mantik.config.environment as _environment
import mantik.utils.mantik_api.client as mantik_api
import mantik.utils.mantik_api.credentials as _credentials
import mantik.utils.mlflow


@pytest.fixture()
def mantik_api_url() -> str:
    return "https://api.test.com"


@pytest.fixture()
def mlflow_tracking_uri() -> str:
    return "https://tracking.test-uri.com"


@pytest.fixture()
def required_env_vars(mlflow_tracking_uri) -> t.Dict[str, str]:
    return {
        _credentials._MANTIK_USERNAME_ENV_VAR: "test-user",
        _credentials._MANTIK_PASSWORD_ENV_VAR: "test-password",
        mantik.utils.mlflow.TRACKING_URI_ENV_VAR: mlflow_tracking_uri,
        # If the env vars for MLflow user/password are set, these are
        # prioritized by MLflow over the token. This leads to an
        # `Unauthorized` error.
        mantik.utils.mlflow.TRACKING_USERNAME_ENV_VAR: "must-be-unset",
        mantik.utils.mlflow.TRACKING_PASSWORD_ENV_VAR: "must-be-unset",
    }


@pytest.fixture()
def token_expiration_date() -> datetime.datetime:
    return datetime.datetime(2022, 1, 1)


@pytest.fixture()
def tmp_dir_as_test_mantik_folder(tmp_path):
    track._MANTIK_FOLDER = pathlib.Path(tmp_path)
    track._MANTIK_TOKEN_FILE = track._MANTIK_FOLDER / "tokens.json"
    return tmp_path


def pytest_configure(config):
    """Remove all MLFLOW_ related environment variables before running any
    test to simplify tests setup."""
    for env in _environment._get_mlflow_env_vars():
        del os.environ[env]


@pytest.fixture
def mock_mantik_api_request(
    env_vars_set, mantik_api_url, expect_raise_if_exception
):
    @contextlib.contextmanager
    def wrapped(
        method: str,
        end_point: str,
        status_code: int,
        json_response: dict,
        expected_error: t.Optional[Exception] = None,
    ) -> None:
        env_vars = {mantik_api._MANTIK_API_URL_ENV_VAR: mantik_api_url}
        with requests_mock.Mocker() as m, env_vars_set(
            env_vars
        ), expect_raise_if_exception(expected_error) as e:
            m.register_uri(
                method=method,
                url=f"{mantik_api_url}{end_point}",
                status_code=status_code,
                json=json_response,
            )
            yield m, e

    return wrapped


@pytest.fixture()
def user_id():
    return "7fc6a4c0-0eb2-4ca0-9e18-c517d7dc92a5"


@pytest.fixture()
def mock_token_sub(mocker, user_id):
    with mocker.patch(
        "mantik.unicore.credentials._get_sub_from_token", return_value=user_id
    ) as mock_sub:
        yield mock_sub


@pytest.fixture
def info_caplog(caplog):
    caplog.set_level(logging.INFO)
    yield caplog


@pytest.fixture
def project_id():
    return uuid.uuid4()



import fastapi

import mantik_compute_backend.app as _app
import mantik_compute_backend.exceptions as _exceptions


def test_create_app():
    docs_url = "/foo"
    app = _app.create_app(
        docs_url=docs_url,
        redoc_url="/bar",
        openapi_url="/baz.json",
    )
    assert isinstance(app, fastapi.FastAPI)
    route_paths = [route.path for route in app.routes]
    assert "/submit/{experiment_id}" in route_paths
    assert docs_url in route_paths
    exception_handlers = app.exception_handlers
    assert _exceptions.unicore_exception_handler in exception_handlers.values()



import abc
import dataclasses
import datetime
import pathlib
import typing as t
import uuid

import fs.base

FilePath = t.Union[str, pathlib.Path]

REMOTE_FS_BASE_PATH_ENV_VAR = "MANTIK_REMOTE_FS_BASE_PATH"
LOCAL_FS_BASE_PATH_ENV_VAR = "MANTIK_LOCAL_FS_BASE_PATH"
REMOTE_FS_TYPE_ENV_VAR = "REMOTE_FS_TYPE"


@dataclasses.dataclass
class FileMeta:
    owner: str
    mode: str
    last_changed: datetime.datetime


@dataclasses.dataclass
class File:
    path: FilePath
    metadata: FileMeta
    is_remote: bool


@dataclasses.dataclass
class Directory(File):
    children: t.List[t.Union["Directory", "File"]]


class AbstractFileService(abc.ABC):
    """
    Abstract class to define methods used for (remote) file handling.

    This interface must be easily implementable with common file transfer
    methods (FTP, boto/S3, GNU filesystem, pathlib + python IO ...).
    """

    def __init__(
        self,
        local_base_path: str = ".",
    ):
        self.relative_local_fs = fs.open_fs(local_base_path)
        self.absolute_local_fs = fs.open_fs("/")

    @abc.abstractmethod
    def list_directory(
        self, target: FilePath
    ) -> t.List[t.Union[Directory, File]]:
        """
        List content of directory.

        Note: bash ls
        """

    @abc.abstractmethod
    def create_directory(self, target: FilePath) -> Directory:
        """
        Make a new directory.

        Note: bash mkdir
        """

    @abc.abstractmethod
    def remove_directory(self, target: FilePath) -> None:
        """Remove a directory.

        Note: bash rm -r
        """

    @abc.abstractmethod
    def copy_directory(
        self,
        source: FilePath,
        target: FilePath,
    ) -> Directory:
        """Copy directory.

        Note: bash cp
        """

    @abc.abstractmethod
    def create_file_if_not_exists(self, target: FilePath) -> File:
        """Create (empty) file if not exists.

        Note: bash touch
        """

    @abc.abstractmethod
    def remove_file(self, target: FilePath) -> None:
        """Remove file or directory.

        Note: bash rm
        """

    @abc.abstractmethod
    def copy_file(
        self,
        source: FilePath,
        target: FilePath,
    ) -> File:
        """Copy file.

        Note: bash cp
        """

    @abc.abstractmethod
    def exists(self, target=FilePath) -> bool:
        """Return if file exists"""

    @property
    @abc.abstractmethod
    def user(self) -> str:
        """Return current user."""

    @abc.abstractmethod
    def change_permissions(
        self, target: FilePath, new_permissions: FileMeta
    ) -> None:
        """Change metadata (permissions) of a file.

        Note: bash chmod
        """

    @classmethod
    @abc.abstractmethod
    def from_env(
        cls, connection_id: t.Optional[uuid.UUID] = None
    ) -> "AbstractFileService":
        """Instantiate with environment variables.

        Credentials are either fetched from mantik api
        or passed in through end vars.
        """

    @classmethod
    @abc.abstractmethod
    def localise_path(self, target: FilePath) -> FilePath:
        """Localize a remote path"""

    @classmethod
    @abc.abstractmethod
    def is_remote(self, target: FilePath) -> bool:
        """Return if a path is remote or not"""

    @abc.abstractmethod
    def _get_remote_fs(self, **nargs) -> fs.base.FS:
        """Return the remote file system"""

    def _get_current_fs_and_relative_filepath(self, path: FilePath):
        current_fs = self._get_fs(path)
        relative_path = (
            self.localise_path(path) if self.is_remote(path) else path
        )

        return current_fs, relative_path

    def _get_fs(self, path: FilePath):
        if self.is_remote(path):
            return self._get_remote_fs(**self._get_fs_nargs(path=path))
        if isinstance(path, pathlib.Path):
            path = path.as_posix()
        if path.startswith("/"):
            return self.absolute_local_fs
        return self.relative_local_fs

    def _get_fs_nargs(self, path: FilePath) -> dict:
        """Extra nargs that might be needed for getting the remote fs"""
        return {}



import pathlib

import pytest

import mantik.config.exceptions as exceptions
import mantik.config.executable as executable


def _create_apptainer_run_command(options: str = "") -> str:
    if options:
        options = f"{options} "
    return f"srun apptainer run {options}image.sif"


class TestApptainer:
    def test_invalid_image_type(self, expect_raise_if_exception):
        expected = (
            "The given Apptainer image type "
            "'unsupported' is not supported, the supported ones are: "
            "'local', 'remote'."
        )
        with expect_raise_if_exception(exceptions.ConfigValidationError()) as e:
            executable.Apptainer(
                path=pathlib.Path("image.sif"),
                type="unsupported",
            )

        result = str(e.value)

        assert result == expected

    @pytest.mark.parametrize(
        ("apptainer", "expected"),
        [
            (
                executable.Apptainer(
                    path=pathlib.Path("/remote/path/to/image.sif"),
                    type="remote",
                ),
                True,
            ),
            (
                executable.Apptainer(
                    path=pathlib.Path("local/path/to/image.sif"),
                    type="local",
                ),
                False,
            ),
        ],
    )
    def test_is_remote(self, apptainer, expected):
        result = apptainer.is_remote

        assert result == expected

    @pytest.mark.parametrize(
        ("apptainer", "absolute_to", "expected"),
        [
            (
                executable.Apptainer(
                    path=pathlib.Path("/remote/path/to/image.sif"),
                    type="remote",
                ),
                pathlib.Path("/path"),
                pathlib.Path("/remote/path/to/image.sif"),
            ),
            (
                executable.Apptainer(
                    path=pathlib.Path("local/path/to/image.sif"),
                    type="local",
                ),
                pathlib.Path("/path"),
                pathlib.Path("/path/local/path/to/image.sif"),
            ),
        ],
    )
    def test_path_as_absolute_to(self, apptainer, absolute_to, expected):
        result = apptainer.path_as_absolute_to(absolute_to)

        assert result == expected

    @pytest.mark.parametrize(
        (
            "environment",
            "expected",
        ),
        [
            # Test Case: Execution command with Apptainer executable
            (
                executable.Apptainer(path=pathlib.Path("image.sif")),
                _create_apptainer_run_command(),
            ),
            # Test Case: Execution command with Apptainer executable
            # and one option as list
            (
                executable.Apptainer(
                    path=pathlib.Path("image.sif"),
                    options=["--B $PWD/data:/data"],
                ),
                _create_apptainer_run_command(options="--B $PWD/data:/data"),
            ),
            # Test Case: Execution command with Apptainer executable
            # and two options
            (
                executable.Apptainer(
                    path=pathlib.Path("image.sif"),
                    options=["--nv", "--B $PWD/data:/data"],
                ),
                _create_apptainer_run_command(
                    options="--nv --B $PWD/data:/data"
                ),
            ),
        ],
    )
    def test_create_run_command(self, environment, expected):
        result = " ".join(
            [environment.get_execution_command(), environment.get_arguments()]
        )

        assert result == expected

    @pytest.mark.parametrize(
        (
            "apptainer_image",
            "expected",
        ),
        [
            (executable.Apptainer(path=pathlib.Path("image.sif")), "image.sif"),
            (
                executable.Apptainer(
                    path=pathlib.Path("/absolute/path/to/image.sif"),
                    type="local",
                ),
                "image.sif",
            ),
            (
                executable.Apptainer(
                    path=pathlib.Path("/absolute/path/to/image.sif"),
                    type="remote",
                ),
                "/absolute/path/to/image.sif",
            ),
        ],
    )
    def test_get_image_path(
        self,
        apptainer_image,
        example_unicore_config,
        expected,
    ):
        assert expected == apptainer_image._get_image_path()


class TestPython:
    def test_path_has_to_be_checked(self):
        python = executable.Python(path=pathlib.Path("/path/to/python/env"))
        expected = False

        result = python.path_has_to_be_checked

        assert result == expected

    def test_create_run_command(self):
        environment = executable.Python(
            path=pathlib.Path("/path/to/python/env")
        )
        expected = "source /path/to/python/env/bin/activate"

        result = environment.get_execution_command()

        assert result == expected



import io
import os
import pathlib
import unittest
import uuid
import zipfile

import pytest
import responses

import mantik.runs.schemas as run_schemas
import mantik.utils.mantik_api.client as mantik_api
import mantik.utils.mantik_api.code_repository as code_api
import mantik.utils.mantik_api.connection as connection
import mantik.utils.mantik_api.experiment_repository as experiment_api
import mantik.utils.mantik_api.user as user

FAKE_API_URL = "https://fake.com"


@pytest.fixture(scope="session")
def sample_experiment_repository_id():
    return uuid.uuid4()


@pytest.fixture(scope="session")
def sample_code_repository_id():
    return uuid.uuid4()


@pytest.fixture(scope="session")
def sample_project_id() -> uuid.UUID:
    return uuid.uuid4()


@pytest.fixture()
def sample_experiment_repository(sample_experiment_repository_id):
    return experiment_api.ExperimentRepository(
        experiment_repository_id=sample_experiment_repository_id,
        mlflow_experiment_id="123",
        name="Name",
        artifact_location="somewhere.com",
    )


@pytest.fixture(scope="session")
def sample_connection_uuid():
    return uuid.uuid4()


@pytest.fixture()
def sample_code_repository(sample_code_repository_id, sample_connection_uuid):
    return code_api.CodeRepository(
        code_repository_id=sample_code_repository_id,
        code_repository_name="Name",
        uri="some/uri.git",
        connection_id=None,
        platform="GitHub",
    )


@pytest.fixture()
def sample_code_repository_with_git_connection(
    sample_code_repository_id, fake_git_connection
):
    return code_api.CodeRepository(
        code_repository_id=sample_code_repository_id,
        code_repository_name="Name",
        uri="some/uri.git",
        connection_id=str(fake_git_connection.connection_id),
        platform="GitHub",
    )


@pytest.fixture()
def fake_git_connection(sample_connection_uuid):
    return connection.Connection(
        connection_id=sample_connection_uuid,
        user=user.User(
            user_id=uuid.UUID("3fa85f64-5717-4562-b3fc-2c963f66afa6"),
            name="test-name",
        ),
        connection_name="git access key",
        connection_provider="GitHub",
        auth_method="token",
        token="auth-122434",
    )


@pytest.fixture(scope="session")
def sample_project_uuid():
    return uuid.uuid4()


@pytest.fixture
def mock_mantik_api_url():
    os.environ[mantik_api._MANTIK_API_URL_ENV_VAR] = FAKE_API_URL
    yield
    os.unsetenv(mantik_api._MANTIK_API_URL_ENV_VAR)


@pytest.fixture
def mock_code_git_connection_response(
    sample_project_uuid,
    fake_user_id_str,
    sample_connection_uuid,
    mock_mantik_api_url,
):
    responses.add(
        responses.GET,
        url=f"{FAKE_API_URL}/users/{fake_user_id_str}/settings/"
        f"connections/{str(sample_connection_uuid)}",
        status=200,
        json={
            "connectionId": str(sample_connection_uuid),
            "user": {"userId": fake_user_id_str, "name": "Kiwi"},
            "connectionName": "KiwiConnection",
            "connectionProvider": "GitHub",
            "authMethod": "Token",
            "token": "token",
        },
    )
    yield


@pytest.fixture()
def mock_get_git_connection(mocker, fake_git_connection):
    with mocker.patch(
        "mantik.utils.mantik_api.connection.get",
        return_value=fake_git_connection,
    ):
        yield


@pytest.fixture()
def mock_get_user_id_from_token(mocker, fake_user_id_str):
    with mocker.patch(
        "mantik.authentication.tokens.get_user_id_from_token",
        return_value=fake_user_id_str,
    ):
        yield


@pytest.fixture(scope="session")
def fake_user_id_str():
    return str(uuid.uuid4())


@pytest.fixture(scope="session")
def sample_run_configuration(
    sample_experiment_repository_id, sample_code_repository_id
) -> run_schemas.RunConfiguration:
    return run_schemas.RunConfiguration(
        name="Sample",
        experiment_repository_id=sample_experiment_repository_id,
        code_repository_id=sample_code_repository_id,
        branch="branch",
        commit="commit",
        data_repository_id=uuid.uuid4(),
        mlflow_mlproject_file_path="some/path/MLProject",
        entry_point="main",
        mlflow_parameters={"output": "hello world"},
        backend_config={},
    )


@pytest.fixture()
def fake_token():
    return "1234"


@pytest.fixture
def resource():
    return (
        pathlib.Path(__file__).parent.parent.parent
        / "resources/test-project/MLproject"
    )


@pytest.fixture
def mock_get_artifacts_url():
    yield unittest.mock.patch(
        "mantik.utils.mantik_api.run.get_download_artifact_url",
        return_value="https://fake/url/artifacts.zip?yada&&yada",
    )


@pytest.fixture
def zipped_file_name():
    return "test_file_name"


@pytest.fixture
def zipped_file_bytes(tmpdir, zipped_file_name, resource):
    def create_zip(input_file, output_zip):
        with zipfile.ZipFile(output_zip, "w") as zipf:
            zipf.write(input_file)
        with open(output_zip, "rb") as zip_file:
            byte_stream = io.BytesIO(zip_file.read())
        return byte_stream

    return create_zip(resource, tmpdir / zipped_file_name)


@pytest.fixture
def mock_get_url(zipped_file_bytes):
    yield unittest.mock.patch("requests.get", return_value=zipped_file_bytes)


@pytest.fixture
def mock_authentication():
    with unittest.mock.patch(
        "mantik.authentication.auth.get_valid_access_token",
        return_value="1234-token",
    ):
        yield "1234-token"



import pytest

import mantik.remote_file_service.s3_file_service as s3_fs


@pytest.mark.parametrize(
    "path,expected",
    [
        ("s3://bucket/resource.item", True),
        ("folder/sub-folder/resource.item", False),
    ],
)
def test_is_remote(path: str, expected: bool) -> None:
    assert s3_fs.S3FileService.is_remote(path) == expected


@pytest.mark.parametrize(
    "path,expected",
    [
        ("s3://bucket-name/resource.item", "resource.item"),
        ("s3://bucket-name", ""),
    ],
)
def test_localise_path(path, expected) -> None:
    assert s3_fs.S3FileService.localise_path(path) == expected


@pytest.mark.parametrize(
    "path,expected",
    [
        ("s3://bucket-name/resource.item", "bucket-name"),
        ("folder/sub-folder/resource.item", "ValueError"),
    ],
)
def test_get_bucket_name(path, expected) -> None:
    if expected == "ValueError":
        with pytest.raises(ValueError):
            s3_fs.S3FileService.get_bucket_name(path)
            return
    else:
        assert s3_fs.S3FileService.get_bucket_name(path) == expected



import abc
import argparse
import contextlib
import dataclasses
import logging
import typing as t

import mlflow


@contextlib.contextmanager
def dry_run_context(*args, **kwargs):
    yield


Context = t.Union[mlflow.start_run, dry_run_context]


def setup_logger(logger: logging.Logger) -> None:
    formatter = logging.Formatter(
        "%(asctime)s.%(msecs)03d %(levelname)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    logger.handlers[0].setFormatter(formatter)


@dataclasses.dataclass
class ResultsBase(abc.ABC):
    mlflow_method: str
    duration: float
    failed: bool

    @abc.abstractmethod
    def to_table_row(self) -> str:
        raise NotImplementedError


def setup_argpaser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="MLflow Tracking Server Stress Tests",
        description="""Runs stress tests against the MLflow Tracking Server.

        Logs 10, 100, 1000, and 10000 and parameters

        1. as individual calls of `mlflow.log_param`
        2. as one call of `mlflow.log_params` with all parameters
        3. as one call of `mlflow.MlflowClient.log_batch` with all parameters

        and prints the result of each test.

        """,
    )
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("-s", "--sleep-time", default=120)
    parser.add_argument("-d", "--dry-run", action="store_true")
    return parser


def get_context(dry_run: bool) -> contextlib.contextmanager:
    return mlflow.start_run if not dry_run else dry_run_context


def log_results(
    results: t.List[ResultsBase],
    columns: t.Iterable[str],
    logger: logging.Logger,
) -> None:
    rows = [result.to_table_row() for result in results]
    header = f"| {' | '.join(columns)} |\n"
    separator = f"| {' | '.join('---' for _ in columns)} |\n"
    table = "".join([header, separator, *rows])
    logger.info("Stress test results: \n%s", table)



import os
import typing as t

import mantik.utils.mlflow as mlflow


def assert_env_var(key: str, expected_value: t.Optional[str]) -> None:
    """Assert given environment variable has given value.

    Parameters
    ----------
    key : str
        Name of the variable.
    expected_value : str or None
        Expected value.

    Raises
    ------
    AssertionError
        If given environment variable
        does not have the expected value.

    """
    value = os.environ.get(key, None)
    assert value == expected_value


def assert_conflicting_mlflow_env_vars_not_set() -> None:
    """
    Assert that the conflicting environemnt variables of mlflow are not set.

    If `MLFLOW_TRACKING_USERNAME` or `MLFLOW_TRACKING_PASSWORD` are set, they
    are preferred by mlflow over `MLFLOW_TRACKING_TOKEN`, and thus, prevent
    the tracking from succeeding.

    """
    assert os.getenv(mlflow.TRACKING_USERNAME_ENV_VAR) is None
    assert os.getenv(mlflow.TRACKING_PASSWORD_ENV_VAR) is None


def assert_correct_tracking_token_env_var_set(expected_access_token) -> None:
    """
    Assert that the correct tracking token environment variable is set.
    """
    assert os.getenv(mlflow.TRACKING_TOKEN_ENV_VAR) == expected_access_token



import uuid

import pytest
import responses

import mantik
import mantik.utils.env as env
import mantik.utils.env_vars as env_vars
import mantik.utils.mantik_api.client as mantik_api_client


@pytest.fixture(scope="session")
def sample_project_id():
    return str(uuid.uuid4())


@pytest.fixture(scope="session")
def sample_run_id():
    return str(uuid.uuid4())


@pytest.fixture(scope="function")
def mock_init_environment(
    sample_run_id, sample_project_id, mantik_api_url
) -> None:
    with env.env_vars_overwrite_temporarily(
        {
            mantik_api_client._MANTIK_API_URL_ENV_VAR: mantik_api_url,
            env_vars.MANTIK_ACCESS_TOKEN_ENV_VAR: "1234",
            env_vars.PROJECT_ID_ENV_VAR: sample_project_id,
            env_vars.MLFLOW_RUN_ID_ENV_VAR: sample_run_id,
        }
    ):
        yield


@pytest.fixture
def mock_bookkeep_infrastructure_api_request(
    mantik_api_url, sample_project_id, sample_run_id
):
    url = (
        f"{mantik_api_url}/projects/{sample_project_id}"
        f"/runs/{sample_run_id}/infrastructure"
    )
    responses.add(
        method="PUT",
        url=url,
        status=200,
    )
    yield url


@responses.activate
def test_init_with_correct_environment_setup_bookkeeps_infrastructure(
    mock_bookkeep_infrastructure_api_request,
    mock_init_environment,
) -> None:
    """
    Given I have project id, run id and in the environment
    When  I call the `mantik.init()` function
    Then  my infrastructure gets bookkept
    And   it gets bookkept on the correct project and run
    """
    mantik.init()
    responses.assert_call_count(
        url=mock_bookkeep_infrastructure_api_request, count=1
    )


def test_init_without_environment_setup_returns_errors() -> None:
    """
    Given I don't have project/run ids in the environment
    When  I call the `mantik.init()` function
    Then  my code crashes
    """
    with pytest.raises(RuntimeError):
        mantik.init()



import pytest

import mantik.remote_file_service.unicore_file_service as unicore_fs


@pytest.mark.parametrize(
    "path,expected",
    [
        ("remote:/path/local.file", True),
        ("/path/local.file", False),
    ],
)
def test_is_remote(path: str, expected: bool) -> None:
    assert unicore_fs.UnicoreFileService.is_remote(path) == expected


@pytest.mark.parametrize(
    "path,expected",
    [
        ("remote:/path/local.file", "/path/local.file"),
        ("remote:/path", "/path"),
    ],
)
def test_localise_path(path, expected) -> None:
    assert unicore_fs.UnicoreFileService.localise_path(path=path) == expected



import unittest.mock

import click.testing
import pytest

import mantik.cli.main as main


@pytest.fixture()
def mock_runs_local_run():
    with unittest.mock.patch("mantik.runs.local.run") as _patch:
        yield _patch


@pytest.fixture()
def mock_get_valid_access_token(fake_token):
    with unittest.mock.patch(
        "mantik.authentication.auth.get_valid_access_token",
        return_value=fake_token,
    ) as _patch:
        yield _patch


def test_trigger_local_run_calls_correct_function_with_correct_arguments(
    mock_get_valid_access_token,
    mock_runs_local_run,
    sample_project_id,
    sample_run_configuration,
    fake_token,
) -> None:
    runner = click.testing.CliRunner()
    result = runner.invoke(
        main.cli,
        [
            "runs",
            "local",
            f"{sample_run_configuration.mlflow_mlproject_file_path}",
            f"--project-id={str(sample_project_id)}",
            f"--name={sample_run_configuration.name}",
            f"--entry-point={sample_run_configuration.entry_point}",
            f"--data-repository-id={str(sample_run_configuration.data_repository_id)}",  # noqa
            f"--experiment-repository-id={str(sample_run_configuration.experiment_repository_id)}",  # noqa
            "--data-target-dir=data",
            f"--code-repository-id={str(sample_run_configuration.code_repository_id)}",  # noqa
            f"--branch={sample_run_configuration.branch}",
            f"--commit={sample_run_configuration.commit}",
            '-Poutput="hello world"',
        ],
    )
    assert result.exit_code == 0
    mock_runs_local_run.assert_called_with(
        data=sample_run_configuration,
        project_id=sample_project_id,
        mantik_token=fake_token,
        data_target_dir="data",
        env_manager="local",
    )



import logging
import typing as t

import requests

import mantik.authentication.tokens as _tokens
import mantik.utils.mantik_api.client as mantik_api
import mantik.utils.mantik_api.credentials as _credentials

logger = logging.getLogger(__name__)

REFRESH_TOKEN_EXPIRED_ERROR_MESSAGE = "Refresh token has expired"
REFRESH_TOKEN_INVALID_ERROR_MESSAGE = "Refresh token is invalid"


def create_tokens(
    credentials: t.Optional[_credentials.Credentials] = None,
) -> _tokens.Tokens:
    logger.info("Creating new tokens")
    return _tokens.Tokens.from_json_response(
        mantik_api.create_tokens(credentials)
    )


def refresh_tokens(
    tokens: _tokens.Tokens,
    credentials: t.Optional[_credentials.Credentials] = None,
) -> _tokens.Tokens:
    """Refresh the tokens.

    Raises
    ------
    RuntimeError
        If MLflow tracking URI environment variable is not set.

    Notes
    -----
    Refreshing a password requires to send the refresh token instead of the
    user's password.

    """
    logger.debug("Refreshing access token")
    try:
        return _tokens.Tokens.from_json_response(
            mantik_api.refresh_tokens(
                refresh_token=tokens.refresh_token,
                access_credentials=credentials,
            ),
            refresh_token=tokens.refresh_token,
        )
    except requests.HTTPError as e:
        if _refresh_token_has_expired(e) or _refresh_token_is_invalid(e):
            logger.debug("Refresh token has expired or is invalid")
            return create_tokens(credentials)
        raise e


def _refresh_token_has_expired(e: requests.HTTPError) -> bool:
    return (
        e.response.status_code == 401
        and REFRESH_TOKEN_EXPIRED_ERROR_MESSAGE in e.response.text
    )


def _refresh_token_is_invalid(e: requests.HTTPError) -> bool:
    return (
        e.response.status_code == 401
        and REFRESH_TOKEN_INVALID_ERROR_MESSAGE in e.response.text
    )



import dataclasses
import datetime
import json
import logging
import pathlib
import typing as t

import jose.jwt

logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class Tokens:
    """Holds AWS Cognito auth tokens."""

    access_token: str
    refresh_token: str
    expires_at: datetime.datetime
    __encoding = "utf-8"

    @classmethod
    def from_json_response(
        cls, response: t.Dict, refresh_token: t.Optional[str] = None
    ) -> "Tokens":
        """Create from JSON response."""
        access_token = response["AccessToken"]
        expires_at = datetime.datetime.fromisoformat(response["ExpiresAt"])
        if refresh_token is None:
            refresh_token = response["RefreshToken"]
        return cls(
            access_token=access_token,
            refresh_token=refresh_token,
            expires_at=expires_at,
        )

    @classmethod
    def from_file(cls, path: pathlib.Path) -> "Tokens":
        """Create from a JSON file."""
        with open(path, encoding=cls.__encoding) as f:
            data = json.load(f, object_hook=_deserialize_datetime)
            return cls(**data)

    @property
    def has_expired(self) -> bool:
        """Return whether the token has expired."""
        tz = self.expires_at.tzinfo
        return self.expires_at < datetime.datetime.now(tz=tz)

    def write_to_file(self, path: pathlib.Path) -> None:
        """Write tokens to file."""
        if not path.parent.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
        logger.debug("Writing tokens %s to file %s", self, path.as_posix())
        with open(path, "w", encoding=self.__encoding) as f:
            data = dataclasses.asdict(self)
            json.dump(
                data,
                f,
                ensure_ascii=False,
                default=_serialize_datetime,
            )


def _deserialize_datetime(
    data: t.Dict[str, str]
) -> t.Dict[str, t.Union[str, datetime.datetime]]:
    expires_at_attribute = "expires_at"
    try:
        value = data[expires_at_attribute]
    except KeyError:
        raise KeyError(f"Token file is missing '{expires_at_attribute}' field")
    data[expires_at_attribute] = datetime.datetime.fromisoformat(value)
    return data


def _serialize_datetime(value: t.Any) -> str:
    if isinstance(value, datetime.datetime):
        return value.isoformat()
    return str(value)


def get_user_id_from_token(token: str) -> str:
    return jose.jwt.get_unverified_claims(token)["sub"]



import json
import pathlib
import typing as t

import yaml

import mantik.config.exceptions as exceptions
import mantik.utils as utils

_ALLOWED_FILE_EXTENSIONS = {"json": [".json"], "yaml": [".yml", ".yaml"]}


def read_config(path: t.Union[pathlib.Path, str]) -> t.Dict:
    if isinstance(path, str):
        path = pathlib.Path(path)
    try:
        if path.suffix in _ALLOWED_FILE_EXTENSIONS["json"]:
            return _read_json_config(path)
        elif path.suffix in _ALLOWED_FILE_EXTENSIONS["yaml"]:
            return _read_yaml_config(path)
    except FileNotFoundError:
        raise exceptions.ConfigValidationError(f"No such config: '{path.name}'")
    allowed_file_types = utils.formatting.iterable_to_string(
        item
        for sublist in _ALLOWED_FILE_EXTENSIONS.values()
        for item in sublist
    )
    raise exceptions.ConfigValidationError(
        f"The given file type {path.suffix!r} "
        "is not supported for the config, "
        "the supported ones are: "
        f"{allowed_file_types}."
    )


def _read_yaml_config(backend_config_path: pathlib.Path) -> t.Dict:
    with open(backend_config_path) as f:
        return yaml.safe_load(f)


def _read_json_config(backend_config_path: pathlib.Path) -> t.Dict:
    with open(backend_config_path) as f:
        return json.load(f)



import io
import os
import pathlib
import tempfile
import typing as t
import zipfile

import mantik.config.core as core
import mantik.utils.unicore.upload as upload_files


def zip_directory_with_exclusion(
    path: pathlib.Path, config: core.Config
) -> t.BinaryIO:
    """Zip given directory."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        with zipfile.ZipFile(tmp_dir + "/zip_tmp", "w") as zipwriter:
            for filepath in upload_files.get_files_to_upload(path, config):
                zip_path = os.path.relpath(filepath, path)
                zipwriter.write(filepath, zip_path)
            zipwriter.close()
        with open(tmp_dir + "/zip_tmp", "rb") as f:
            content = f.read()
        return io.BytesIO(content)



"""
Example Access Token:
eyJraWQiOiJqWnd6VW0xOUtOOWplK1ZpVjBLeEdvUW5kemZQXC9FVTBrVzlONzA0NEVWVT0iLCJhbGciOiJSUzI1NiJ9.eyJzdWIiOiIwYmRkMzQxYy0yODFmLTRlNTEtYWEzNy01MDI5MTU3NDkwNDMiLCJpc3MiOiJodHRwczpcL1wvY29nbml0by1pZHAuZXUtY2VudHJhbC0xLmFtYXpvbmF3cy5jb21cL2V1LWNlbnRyYWwtMV9zOHBIMkt4T1AiLCJjbGllbnRfaWQiOiIydmRmOTh2cWVwb2VlMTRocXBjODQ4NmFsZSIsIm9yaWdpbl9qdGkiOiIyMzIyYmYzYi0xOTI2LTRjM2UtYmI5ZS05N2YzMmM0YTA2NWMiLCJldmVudF9pZCI6ImFkNzQ5MzQ2LTlkMGMtNDYzOC04MzMxLWViMGMyNjY3OTMyZiIsInRva2VuX3VzZSI6ImFjY2VzcyIsInNjb3BlIjoiYXdzLmNvZ25pdG8uc2lnbmluLnVzZXIuYWRtaW4iLCJhdXRoX3RpbWUiOjE2NTcwOTgxMTgsImV4cCI6MTY1NzEwMTcxOCwiaWF0IjoxNjU3MDk4MTE4LCJqdGkiOiIxZjc2OGFkNy0wNzI3LTQ2MWYtYTJhMC00MjQzNzM5ZTZjM2EiLCJ1c2VybmFtZSI6ImZhYmlhbiJ9.g9Gz-AzAgKhVxec_faSPnhVbBFEeqeg4XaGtWPe9TgusiFkWQmQp6nBMkZAPEOIxcYJd7MFyFtv6vhPsz6PgmrXQN-FqHZ4eEvmbjJEZej-9iu535Rft5BtfIpTswqdBAUggBd9hcCqgCJgCn8nJ9PpYledPskI_7uzTxaZOkhtsoeMfr6BT8gpjHoN0GjKWN9FBLeqtN-miI_FoF4OxKfO9hPzeF0n89MkR85FNPsB8FpkwEMZe7D4fVCBreZxrc9vA8kecU9_1D2AjPujODndKn-E5tXfSufrKK2Fj7JJ51F_v1Gk8BFe6fx50dxi3-smSm0VxU7nq7MDf8L9UbQ  # noqa

Corresponding Refresh Token:
eyJjdHkiOiJKV1QiLCJlbmMiOiJBMjU2R0NNIiwiYWxnIjoiUlNBLU9BRVAifQ.Ukk0wBed6TUAIxTrUpxCNwjiglcHc2YNaH0wAsZFZSDrywYn3H0gu-vfPEp4E6z47KPEuJiQBFsCQTXZLmYS3GCJQrw3EpV8QwL5-gdqdgseCrsYjGBgStkRNTP-QqH8_EbJb5L6j0u5qYMn6MWjkCGpK_5A4N7TQ4eLCdPMxMOWTAJRJs6pvc0birVtTq6qmAAjLIVlGMz1b2DwQ89qIE950AzbSJ3uSiAOU0Q9EgEghNAgsRg6b7So202xqkwI98FjgovwG6VhvvEFlOUieH0ZTF7tQSj3ATyb33QlhSh21wjddLd_T1pomPEML5Gj7PslUA_4hxYScksVvoF4Qg.ktbthM0IS218Qqrl.d8mkGvTo9c-PYDQ1Aqd3xFwuVvi_2m0CNOPAjmyWwQQ1pFcWMbVuTIz2h6fP3OITYdHxL2TlAMpEeuE6lKJzI_dRv8YEEii-rnSLp_2Y8QZekT-b8ZLt6ByQqalVIMVXxhbooABP8qQjEPXxd4gsWSTrvgbI94ENlXb5SrqQl2X4NN5RFy_Ag6BhOZyZzc98Xa05kSBAn4EDIrAoxPmFprBMy8EUZuJ2z7WyAOuQAkH2Au14DcD8VpsZ3xgUBoWQ62rNjZovXnQLoLn1kMnWrPW51clYp0-9D6NOqVp7Bpm7UdID65hGf56o7rdGE1bM_X94DeeeDD94DNVTdy9jdxsL4oW216QPLS-p7NrDJVgDuIqWN_dJUzWJRLuQb5it_KTdGp5-ddWVtYqn0LUxZ0k0cn8sES-0D2o-DYwEcUHx3e9fbCXymdmKkMcDCB2pvCVpM7C_BhcIUKpHZ-7Z4xVncLKD2fd-PCzGULHwdrHnpFuqSJbDtURR9DO8RDhrQIy3l7jVB9CPXF41GycG5G_S9hEMD_C4wxFYKNTsr2zlpspCsQk22PtONrqkNgnnqglPSaBX6MMLzRztMARD2VblxRc4udt6X7cSXmtho1sFRpQKpoQpagHhG0adiXnb-6k9LI4-mFRJI4qFL_DNqcgPn8Bqhdj_4f5g4vF_LGkTSLgunel71DcXmY3W7TpjFUgvwWs8jc7F1W4Pe2xAXCdLvf1EPA2bIXEsMBv-Q89DNHG-VIsR_MzkbYUioJpRBCjWOpLGaNVC4fQ_lkBuBYmvZP7SU7ZR2AJFvT7DVHvUUAPhz_4dprwVV4Znq_r0cZDDmyVIi9QOZxDB17O2rVUCadtPNYU07UF-BE-3IjdI5VCotzf4PQnN486HneMDbb-hh2cD0ZNeEiGUuFrrwHNjSCL3NsSKoLbAOkQbwK6soUbxMk3x-ooS-gPsxbNh9a7b9qtwxH5nOFEKPRcFFehKsdKYB2g4CeuGRwVwgOqXshNojZEuRMt4x0qOrZsUBlcvdtfnuagqquaNGHRxnZ1-cYlKszxZ38q-jC6jKMmbqTb_DUDFgv3eYit7uB6OOLO8gW9G9fLYeRNpeOR4x-zS7pTClvoGGvYtjAPJAvNAxMmXNe4pXrX1mNTrHUjv8uThGoMSHzF1AwHvZLgtO6xj8GbgdnWNr2fVKP_bZczNSXRLYSQ03c65UFI4r1RPde-Zt2zIRJgywxBkgKZodCCtQ2BRlvKGr4KpUFh530nZeJis0or1BirrA7rr3u--9VMcAA6RSaaZdd8N.KwiXuNdKy9Jn2D8uapwVJg  # noqa


Header:
{
  "kid": "jZwzUm19KN9je+ViV0KxGoQndzfP/EU0kW9N7044EVU=",
  "alg": "RS256"
}

Payload:
{
  "sub": "0bdd341c-281f-4e51-aa37-502915749043",
  "iss": "https://cognito-idp.eu-central-1.amazonaws.com/eu-central-1_s8pH2KxOP",  # noqa
  "client_id": "2vdf98vqepoee14hqpc8486ale",
  "origin_jti": "2322bf3b-1926-4c3e-bb9e-97f32c4a065c",
  "event_id": "ad749346-9d0c-4638-8331-eb0c2667932f",
  "token_use": "access",
  "scope": "aws.cognito.signin.user.admin",
  "auth_time": 1657098118,
  "exp": 1657101718,
  "iat": 1657098118,
  "jti": "1f768ad7-0727-461f-a2a0-4243739e6c3a",
  "username": "fabian"
}

Issuer jwks.json file:
{
  "keys":[
    {
      "alg":"RS256",
      "e":"AQAB",
      "kid":"fG9ZI3zslGG8GH308l93bYedOYvON1Kqj/FeCmXIAZM=",
      "kty":"RSA",
      "n":"tC4LJ_vPuCf4qFkk6wXYpUb_qfnT17lnXMYQfTV3MZxXE7z6oG3SffjwYRw2RcFVO1QCIs_Lqc_03fLQB5fQp8_nQtZYTXeTGmO_6WyfcUmSS_GmISnbGbmaOQJCr8dSt06Xd3L4oqg47A5Zy7JesSP7AiVpZk-7o0Zi-S-aIpDr2fZa5bUar0hrJdETKuPMY2Kio9HY1VDGqAnLdtkhvxEYB3jNa3XLAJ4p2ZEMakJHhsIm84iGXBuxo4S1oNL1O8RP9awN6dr_5YuYLcENPOZ_OBb9fes9goAVedhfdflGB5swWVGNUEMBB8ElgFKC4IxELNbJ7XopzMCVTpJX3Q",  # noqa
      "use":"sig"
    },
    {
      "alg":"RS256",
      "e":"AQAB",
      "kid":"jZwzUm19KN9je+ViV0KxGoQndzfP/EU0kW9N7044EVU=",
      "kty":"RSA",
      "n":"rP7xiJVMB1xnutn9QX0cD0CzoVx64zWUyeOo4c0Ed1I6Mu0R0Borwqoh_u6cJd7PBYlnTme4k_ueEDm7Z4w-m-89igM5_0oo0jionmFwCuWmShjoH1dJ2J6lb9v52jiVzQV2rIqlWx2Gkq3IuwgrDIg_11BY1X0lyGhuygncp3z-tRUmnB3kcdkO_gddUyI8Uhb20g1ypeTwx-FzWGPVOXAMaAeDMOak33pWzVVzR3zAvrcw6BHVgr3gwU7nPxq5TlD_d8lhYI7NvS16aML_OQsBvoJCCXAxf4j9g3MicUWAFSysLRaTCAkxHbrKOvFdEJ9k64G3sdbCTJf5J3MAWQ",  # noqa
      "use":"sig"
    }
  ]
}
"""
import datetime
import typing as t

import tokens as tokens
import tokens.cognito as _cognito
import tokens.jwks as _jwks

FAKE_JWT = "eyJraWQiOiJqWnd6VW0xOUtOOWplK1ZpVjBLeEdvUW5kemZQXC9FVTBrVzlONzA0NEVWVT0iLCJhbGciOiJSUzI1NiJ9.eyJzdWIiOiIwYmRkMzQxYy0yODFmLTRlNTEtYWEzNy01MDI5MTU3NDkwNDMiLCJpc3MiOiJodHRwczpcL1wvY29nbml0by1pZHAuZXUtY2VudHJhbC0xLmFtYXpvbmF3cy5jb21cL2V1LWNlbnRyYWwtMV9zOHBIMkt4T1AiLCJjbGllbnRfaWQiOiIydmRmOTh2cWVwb2VlMTRocXBjODQ4NmFsZSIsIm9yaWdpbl9qdGkiOiIyMzIyYmYzYi0xOTI2LTRjM2UtYmI5ZS05N2YzMmM0YTA2NWMiLCJldmVudF9pZCI6ImFkNzQ5MzQ2LTlkMGMtNDYzOC04MzMxLWViMGMyNjY3OTMyZiIsInRva2VuX3VzZSI6ImFjY2VzcyIsInNjb3BlIjoiYXdzLmNvZ25pdG8uc2lnbmluLnVzZXIuYWRtaW4iLCJhdXRoX3RpbWUiOjE2NTcwOTgxMTgsImV4cCI6MTY1NzEwMTcxOCwiaWF0IjoxNjU3MDk4MTE4LCJqdGkiOiIxZjc2OGFkNy0wNzI3LTQ2MWYtYTJhMC00MjQzNzM5ZTZjM2EiLCJ1c2VybmFtZSI6ImZhYmlhbiJ9.g9Gz-AzAgKhVxec_faSPnhVbBFEeqeg4XaGtWPe9TgusiFkWQmQp6nBMkZAPEOIxcYJd7MFyFtv6vhPsz6PgmrXQN-FqHZ4eEvmbjJEZej-9iu535Rft5BtfIpTswqdBAUggBd9hcCqgCJgCn8nJ9PpYledPskI_7uzTxaZOkhtsoeMfr6BT8gpjHoN0GjKWN9FBLeqtN-miI_FoF4OxKfO9hPzeF0n89MkR85FNPsB8FpkwEMZe7D4fVCBreZxrc9vA8kecU9_1D2AjPujODndKn-E5tXfSufrKK2Fj7JJ51F_v1Gk8BFe6fx50dxi3-smSm0VxU7nq7MDf8L9UbQ"  # noqa
FAKE_JWT_INVALID_SIGNATURE = "eyJraWQiOiJqWnd6VW0xOUtOOWplK1ZpVjBLeEdvUW5kemZQXC9FVTBrVzlONzA0NEVWVT0iLCJhbGciOiJSUzI1NiJ9.eyJzdWIiOiIwYmRkMzQxYy0yODFmLTRlNTEtYWEzNy01MDI5MTU3NDkwNDMiLCJpc3MiOiJodHRwczpcL1wvY29nbml0by1pZHAuZXUtY2VudHJhbC0xLmFtYXpvbmF3cy5jb21cL2V1LWNlbnRyYWwtMV9zOHBIMkt4T1AiLCJjbGllbnRfaWQiOiIydmRmOTh2cWVwb2VlMTRocXBjODQ4NmFsZSIsIm9yaWdpbl9qdGkiOiIyMzIyYmYzYi0xOTI2LTRjM2UtYmI5ZS05N2YzMmM0YTA2NWMiLCJldmVudF9pZCI6ImFkNzQ5MzQ2LTlkMGMtNDYzOC04MzMxLWViMGMyNjY3OTMyZiIsInRva2VuX3VzZSI6ImFjY2VzcyIsInNjb3BlIjoiYXdzLmNvZ25pdG8uc2lnbmluLnVzZXIuYWRtaW4iLCJhdXRoX3RpbWUiOjE2NTcwOTgxMTgsImV4cCI6MTY1NzEwMTcxOCwiaWF0IjoxNjU3MDk4MTE4LCJqdGkiOiIxZjc2OGFkNy0wNzI3LTQ2MWYtYTJhMC00MjQzNzM5ZTZjM2EiLCJ1c2VybmFtZSI6ImZhYmlhbiJ9.G9Gz-AzAgKhVxec_faSPnhVbBFEEqeg4XaGtWPe9TgUsiFkWQmQp6nBMkZAPEOIxcYJd7MFyFtv6vhPsZ6PgmrXQN-FqHZ4eEvmbjJEZej-9iu535Rft5BtfIpTswqdBAUggbd9hcCqgCJgCn8nJ9PpYledPskI_1uzTxaZOkhtsoeMfr6BT0gpjHoN0GjKWN9FBLeqtN-LiI_FoF4OxKfO9hPzeF0n89MkR85FNPsB8FpkwEMZe7D4fVCBreZxrc9vA8kecU9_1D2AjPujODndKn-E5tXfSufrKK2Fj7JJ51F_v1Gk8BFe6fx50dxi3-smSm0VxU7nq7MDf8L9UbQ"  # noqa
FAKE_JWT_INVALID_CLIENT_ID = "eyJraWQiOiJqWnd6VW0xOUtOOWplK1ZpVjBLeEdvUW5kemZQL0VVMGtXOU43MDQ0RVZVPSIsImFsZyI6IlJTMjU2In0.eyJzdWIiOiIwYmRkMzQxYy0yODFmLTRlNTEtYWEzNy01MDI5MTU3NDkwNDMiLCJpc3MiOiJodHRwczovL2NvZ25pdG8taWRwLmV1LWNlbnRyYWwtMS5hbWF6b25hd3MuY29tL2V1LWNlbnRyYWwtMV9zOHBIMkt4T1AiLCJjbGllbnRfaWQiOiIydmRmOTh2cWVwb0VlMTRocVBjODQ4NmFsZSIsIm9yaWdpbl9qdGkiOiIyMzIyYmYzYi0xOTI2LTRjM2UtYmI5ZS05N2YzMmM0YTA2NWMiLCJldmVudF9pZCI6ImFkNzQ5MzQ2LTlkMGMtNDYzOC04MzMxLWViMGMyNjY3OTMyZiIsInRva2VuX3VzZSI6ImFjY2VzcyIsInNjb3BlIjoiYXdzLmNvZ25pdG8uc2lnbmluLnVzZXIuYWRtaW4iLCJhdXRoX3RpbWUiOjE2NTcwOTgxMTgsImV4cCI6MTY1NzEwMTcxOCwiaWF0IjoxNjU3MDk4MTE4LCJqdGkiOiIxZjc2OGFkNy0wNzI3LTQ2MWYtYTJhMC00MjQzNzM5ZTZjM2EiLCJ1c2VybmFtZSI6ImZhYmlhbiJ9.LbsYIONdeeSWzsgUirHcbGGl7yySIXN5WLHWNpn4GAfXTTr2_HUJli_0Cyi6nf1sJuRJL4o8r_MXwzE0rvKjXkfG9z_jCQEGu8htdmhP6ubia6AWy90uN7ZDzSRmv1tUN7Nd3_vQuqFmfIO_kyl8y0QgmDn4bNu1rKL4iZyAmwS1MGo2JUiZJH62UZA79Iv94EzrBluQ-ePq94PEzJ3rsny-6iMFDgu9WYTxrpkhix4PaNU6DHXBFVEV2VOZ8n9MOOMCmeXpmQR7iauQNRilHdAXEjg7fIa3eaxeShL5ZkRJxTc_Eod_4LY_NnN-zyu66lYQsJNG-rVv3Qt20_ufWQ"  # noqa
FAKE_JWT_INVALID_ISSUER = "eyJraWQiOiJqWnd6VW0xOUtOOWplK1ZpVjBLeEdvUW5kemZQL0VVMGtXOU43MDQ0RVZVPSIsImFsZyI6IlJTMjU2In0.eyJzdWIiOiIwYmRkMzQxYy0yODFmLTRlNTEtYWEzNy01MDI5MTU3NDkwNDMiLCJpc3MiOiJodHRwczovL2NvZ25pdG8taWRwLmV1LWNlbnRyYWwtMS5hbWF6b25hd3MuY29tL2V1LWNlbnRyYWwtMV9zOHBIMkt4b3AiLCJjbGllbnRfaWQiOiIydmRmOTh2cWVwb2VlMTRocXBjODQ4NmFsZSIsIm9yaWdpbl9qdGkiOiIyMzIyYmYzYi0xOTI2LTRjM2UtYmI5ZS05N2YzMmM0YTA2NWMiLCJldmVudF9pZCI6ImFkNzQ5MzQ2LTlkMGMtNDYzOC04MzMxLWViMGMyNjY3OTMyZiIsInRva2VuX3VzZSI6ImFjY2VzcyIsInNjb3BlIjoiYXdzLmNvZ25pdG8uc2lnbmluLnVzZXIuYWRtaW4iLCJhdXRoX3RpbWUiOjE2NTcwOTgxMTgsImV4cCI6MTY1NzEwMTcxOCwiaWF0IjoxNjU3MDk4MTE4LCJqdGkiOiIxZjc2OGFkNy0wNzI3LTQ2MWYtYTJhMC00MjQzNzM5ZTZjM2EiLCJ1c2VybmFtZSI6ImZhYmlhbiJ9.Xjozn1LDc75O2kEIoeVO79ThQ7WwqivgFwaLEVV32BzTNnk8Ic-8nBYmxjyS6kc7yVbO0CIKbYD5LmcFUdPbBvKLvDGXuFUHXVuabjbeT92U8M9CA5G-1SsSykklDJz8GEV0wQfCAXsa4W3XsvE9DmjQP9fl_kleRl1pzW9t4QtXVp1MvUNBsvB-VJHXMsBjtzYm9ZzIgKJAD013rkFZcbyYrNa3cf7aWgoCJ7-rUEknUyphmmBnGoFlBdWsswVReqHF75vmZcQ3xNpFE6J3Z_1piHoxCrtPMprMF5RgGRgMRYb8VzANmOeR1yfAIA_A7a13_lLcimWCSV9KYPM2mg"  # noqa
FAKE_JWT_INVALID_TOKEN_TYPE = "eyJraWQiOiJqWnd6VW0xOUtOOWplK1ZpVjBLeEdvUW5kemZQL0VVMGtXOU43MDQ0RVZVPSIsImFsZyI6IlJTMjU2In0.eyJzdWIiOiIwYmRkMzQxYy0yODFmLTRlNTEtYWEzNy01MDI5MTU3NDkwNDMiLCJpc3MiOiJodHRwczovL2NvZ25pdG8taWRwLmV1LWNlbnRyYWwtMS5hbWF6b25hd3MuY29tL2V1LWNlbnRyYWwtMV9zOHBIMkt4T1AiLCJjbGllbnRfaWQiOiIydmRmOTh2cWVwb2VlMTRocXBjODQ4NmFsZSIsIm9yaWdpbl9qdGkiOiIyMzIyYmYzYi0xOTI2LTRjM2UtYmI5ZS05N2YzMmM0YTA2NWMiLCJldmVudF9pZCI6ImFkNzQ5MzQ2LTlkMGMtNDYzOC04MzMxLWViMGMyNjY3OTMyZiIsInRva2VuX3VzZSI6ImlkIiwic2NvcGUiOiJhd3MuY29nbml0by5zaWduaW4udXNlci5hZG1pbiIsImF1dGhfdGltZSI6MTY1NzA5ODExOCwiZXhwIjoxNjU3MTAxNzE4LCJpYXQiOjE2NTcwOTgxMTgsImp0aSI6IjFmNzY4YWQ3LTA3MjctNDYxZi1hMmEwLTQyNDM3MzllNmMzYSIsInVzZXJuYW1lIjoiZmFiaWFuIn0.U2JWar-tC46W4GDb5TGqIbQtE2KSoxghS8mItWCfoH1YrEkwSi_P2Ujjitamho82aGKXzWJlJseDwEO5-sPHkDKSkWlHqGxm5NKtxx7NZDZhoOD3DO2RPDto8HX8YnqDaxH0h3IMrF59KD_IU-4wtLg33jmajFnejFuifHCDvQyiqGFD88KiMbjQ8bTvoTn9w1INKYfUWqrWWLbdE6XZCZOlP-v4Q49EKNnQooKqFWASCV54a8mgfjGkRm_My2mmkmqpCcjzeQiA2rtKC_zARUTyEBngklNJmcrhS-5_hlG1dtQh71Isz2cmTTZyzA32KG-CQVd7lj8GMSbZvpE5Tg"  # noqa

FAKE_COGNITO_REGION = "eu-central-1"
FAKE_COGNITO_USER_POOL_ID = "eu-central-1_s8pH2KxOP"
FAKE_COGNITO_APP_CLIENT_ID = "2vdf98vqepoee14hqpc8486ale"
FAKE_COGNITO_APP_CLIENT_SECRET = (
    "1cvd3babj7lriqnr2i8d7r16kpfftfk1jpf6d6phutj9aa5l4qb6"
)

ISSUED_AT = datetime.datetime(2022, 7, 6, 9, 1, 58)
EXPIRES_AT = datetime.datetime(2022, 7, 6, 10, 1, 58)
NON_EXPIRED_DATETIME = ISSUED_AT + datetime.timedelta(minutes=10)
EXPIRED_DATETIME = EXPIRES_AT + datetime.timedelta(minutes=1)


class FakeJWKS(tokens.jwks.JWKS):
    @classmethod
    def create(cls) -> "FakeJWKS":
        fake_keys = [
            {
                "alg": "RS256",
                "e": "AQAB",
                "kid": "fG9ZI3zslGG8GH308l93bYedOYvON1Kqj/FeCmXIAZM=",
                "kty": "RSA",
                "n": "tC4LJ_vPuCf4qFkk6wXYpUb_qfnT17lnXMYQfTV3MZxXE7z6oG3SffjwYRw2RcFVO1QCIs_Lqc_03fLQB5fQp8_nQtZYTXeTGmO_6WyfcUmSS_GmISnbGbmaOQJCr8dSt06Xd3L4oqg47A5Zy7JesSP7AiVpZk-7o0Zi-S-aIpDr2fZa5bUar0hrJdETKuPMY2Kio9HY1VDGqAnLdtkhvxEYB3jNa3XLAJ4p2ZEMakJHhsIm84iGXBuxo4S1oNL1O8RP9awN6dr_5YuYLcENPOZ_OBb9fes9goAVedhfdflGB5swWVGNUEMBB8ElgFKC4IxELNbJ7XopzMCVTpJX3Q",  # noqa
                "use": "sig",
            },
            {
                "alg": "RS256",
                "e": "AQAB",
                "kid": "jZwzUm19KN9je+ViV0KxGoQndzfP/EU0kW9N7044EVU=",
                "kty": "RSA",
                "n": "rP7xiJVMB1xnutn9QX0cD0CzoVx64zWUyeOo4c0Ed1I6Mu0R0Borwqoh_u6cJd7PBYlnTme4k_ueEDm7Z4w-m-89igM5_0oo0jionmFwCuWmShjoH1dJ2J6lb9v52jiVzQV2rIqlWx2Gkq3IuwgrDIg_11BY1X0lyGhuygncp3z-tRUmnB3kcdkO_gddUyI8Uhb20g1ypeTwx-FzWGPVOXAMaAeDMOak33pWzVVzR3zAvrcw6BHVgr3gwU7nPxq5TlD_d8lhYI7NvS16aML_OQsBvoJCCXAxf4j9g3MicUWAFSysLRaTCAkxHbrKOvFdEJ9k64G3sdbCTJf5J3MAWQ",  # noqa
                "use": "sig",
            },
        ]
        return cls(keys=fake_keys)


class FakeTokenVerifier(tokens.verifier.TokenVerifier):
    def __init__(
        self,
        cognito: t.Optional[_cognito.client.Properties] = None,
        jwks: t.Optional[_jwks.JWKS] = None,
        secret_required: t.Optional[bool] = None,
    ):
        self._ensure_secret_required_parameter_is_set(secret_required)

    def verify_token(self, token: str) -> None:
        if not token == "test-valid-token":
            raise tokens.exceptions.VerificationFailedException("Invalid token")



import mantik.cli
import mantik.data_repository
import mantik.remote_file_service
import mantik.utils
from mantik.data_repository.data_repository import data_download
from mantik.init import init
from mantik.runs.local import start_local_run
from mantik.runs.remote import submit_run
from mantik.tracking.track import init_tracking



"""Fakes for testing cognito interactions.

Reference:
https://docs.aws.amazon.com/code-samples/latest/catalog/python-test_tools-cognito_idp_stubber.py.html  # noqa

"""
import dataclasses
import typing as t

import boto3
import botocore.session
import botocore.stub


def patch_boto_client(monkeypatch, json_response):
    def create_fake_client(*args, **kwargs):
        return FakeCognitoClient(json_response)

    monkeypatch.setattr(
        boto3,
        "client",
        create_fake_client,
    )


@dataclasses.dataclass
class FakeCognitoClient:
    json_response: t.Dict
    _client = boto3.client("cognito-idp", region_name="test-region")

    def initiate_auth(
        self, ClientId: str, AuthFlow: str, AuthParameters: t.Dict
    ) -> t.Dict:
        """Fake initiate_auth method and raise error if response is one.

        Notes
        -----
        The boto3.client.exceptions classes take two arguments:
            1. json response (`dict`), which is being parsed by the class.
            2. An `operation_name`, which does not seem to be used
               for the error message.

        """
        if "Error" in self.json_response:
            code = self.json_response["Error"]["Code"]
            if code == "NotAuthorizedException":
                raise self._client.exceptions.NotAuthorizedException(
                    self.json_response, ""
                )
            elif code == "UserNotFoundException":
                raise self._client.exceptions.UserNotFoundException(
                    self.json_response, ""
                )
            raise RuntimeError("Unknown error code in response")
        return self.json_response

    def close(self) -> None:
        pass

    @property
    def exceptions(self):
        return self._client.exceptions


class BotoClientPatcher(botocore.stub.Stubber):
    def __init__(self, client: t.Type[boto3.client], use_stubs: bool = True):
        """
        Initializes the object with a specific client and configures it for
        stubbing or AWS passthrough.

        Parameters
        ----------
        client : boto3.client
            A Boto 3 service client.
        use_stubs : bool, default=True.
            When True, use stubs to intercept requests.
            Otherwise, pass requests through to AWS.
        """
        self.use_stubs = use_stubs
        self.region_name = client.meta.region_name
        self.client = client
        if self.use_stubs:
            super().__init__(client)
        else:
            self.client = client

    @property
    def exceptions(self):
        return self.client.exceptions

    def add_response(self, method: str, service_response, expected_params=None):
        """Add response to a method.

        Parameters
        ----------
        method : str
            Name of the client method to stub.
        service_response : dict
            Response data.
        expected_params: : dict
            A dictionary of the expected parameters to
            be called for the provided service response. The parameters match
            the names of keyword arguments passed to that client call. If
            any of the parameters differ a ``StubResponseError`` is thrown.
            You can use stub.ANY to indicate a particular parameter to ignore
            in validation. stub.ANY is only valid for top level params.

        """

        if self.use_stubs:
            super().add_response(method, service_response, expected_params)

    def add_client_error(
        self,
        method,
        service_error_code="",
        service_message="",
        http_status_code=400,
        service_error_meta=None,
        expected_params=None,
        response_meta=None,
    ):
        """Adds a `ClientError` to the response queue.

        Parameters
        ----------
        method : str
            The name of the service method to return the error on.
        service_error_code : str
            The service error code to return.
            E.g. ``NoSuchBucket``.
        service_message : str
            The service message to return.
            E.g. 'The specified bucket does not exist.'
        http_status_code : int
            The HTTP status code to return
            E.g. 404, etc.
        service_error_meta : dict
            Additional keys to be added to the service error.
        expected_params : dict
            A dictionary of the expected parameters to
            be called for the provided service response. The parameters match
            the names of keyword arguments passed to that client call. If
            any of the parameters differ a ``StubResponseError`` is thrown.
            You can use stub.ANY to indicate a particular parameter to ignore
            in validation.
        response_meta : dict
            Additional keys to be added to the response's ResponseMetadata.
        modeled_fields : dict
            Additional keys to be added to the response based on fields that
            are modeled for the particular error code. These keys will be
            validated against the particular error shape designated by the
            error code.

        """
        if self.use_stubs:
            super().add_client_error(
                method,
                service_error_code,
                service_message,
                http_status_code,
                service_error_meta,
                expected_params,
                response_meta,
            )

    def assert_no_pending_responses(self):
        """When using stubs, verify that no more responses are in the queue."""
        if self.use_stubs:
            super().assert_no_pending_responses()

    def _stub_bifurcator(
        self,
        method: str,
        expected_params: t.Dict = None,
        response: t.Dict = None,
        error_code: str = None,
        error_message: str = None,
    ):
        if expected_params is None:
            expected_params = {}
        if response is None:
            response = {}
        if error_code is None:
            self.add_response(
                method,
                expected_params=expected_params,
                service_response=response,
            )
        else:
            self.add_client_error(
                method,
                expected_params=expected_params,
                service_error_code=error_code,
                service_message=error_message,
            )


class CognitoClientPatcher(BotoClientPatcher):
    def __init__(self, use_stubs: bool = True):
        """Initializes the object with a specific client and configures it for
        patching or AWS passthrough.

        Parameters
        ----------
        use_stubs : bool, default=True.
            When True, use stubs to intercept requests.
            Otherwise, pass requests through to AWS.

        """
        client = botocore.session.get_session().create_client(
            "cognito-idp", region_name="test-region"
        )
        super().__init__(client, use_stubs=use_stubs)

    def patch_initiate_auth_get_token(
        self,
        client_id: str,
        user_name: str,
        password: str,
        secret_hash: str,
        response: t.Dict,
        error_code: str = None,
        error_message: str = None,
    ):
        expected_params = {
            "ClientId": client_id,
            "AuthFlow": "USER_PASSWORD_AUTH",
            "AuthParameters": {
                "USERNAME": user_name,
                "PASSWORD": password,
                "SECRET_HASH": secret_hash,
            },
        }
        self._stub_bifurcator(
            method="initiate_auth",
            expected_params=expected_params,
            response=response,
            error_code=error_code,
            error_message=error_message,
        )

    def patch_initiate_auth_refresh_token(
        self,
        client_id: str,
        refresh_token: str,
        secret_hash: str,
        response: t.Dict,
        error_code: str = None,
        error_message: str = None,
    ):
        expected_params = {
            "ClientId": client_id,
            "AuthFlow": "REFRESH_TOKEN_AUTH",
            "AuthParameters": {
                "REFRESH_TOKEN": refresh_token,
                "SECRET_HASH": secret_hash,
            },
        }
        self._stub_bifurcator(
            method="initiate_auth",
            expected_params=expected_params,
            response=response,
            error_code=error_code,
            error_message=error_message,
        )



import logging

logger = logging.getLogger(__name__)


def set_verbose_logging(ctx, param, value) -> None:  # noqa
    """
    Callback function to set log level from a --verbose flag.

    Note: Unused arguments are part of the click callback signature and
    must be present.
    """
    if value:
        logger.root.setLevel(level=logging.DEBUG)
        logger.debug("Using logging level DEBUG")
    else:
        logger.root.setLevel(level=logging.WARNING)
        logging.basicConfig(level=logging.WARNING)



import os
import typing as t
import uuid

import mantik.authentication.auth as auth
import mantik.utils.env_vars as env_vars
import mantik.utils.mantik_api as mantik_api


def init(
    project_id: t.Optional[uuid.UUID] = None,
    run_id: t.Optional[str] = None,
    mantik_access_token: t.Optional[str] = None,
) -> None:
    """Initialise mantik for run bookkeeping

    Including:
    - bookkeep run infrastructure
    """
    try:
        project_id = project_id or os.environ[env_vars.PROJECT_ID_ENV_VAR]
        run_id = (
            run_id
            or os.getenv(env_vars.RUN_ID_ENV_VAR)
            or os.environ[env_vars.MLFLOW_RUN_ID_ENV_VAR]
        )
    except KeyError as e:
        raise RuntimeError(f"mantik.init() requires environment variable {e}")

    mantik_access_token = (
        mantik_access_token
        or os.getenv(env_vars.MANTIK_ACCESS_TOKEN_ENV_VAR)
        or auth.get_valid_access_token()
    )

    mantik_api.run.update_run_infrastructure(
        project_id=project_id,
        run_id=uuid.UUID(run_id),
        token=mantik_access_token,
        infrastructure=mantik_api.run.RunInfrastructure.from_system(),
    )



import pytest

import mantik.config as config


class TestResources:
    def test_to_job_script(self):
        expected = (
            "#SBATCH --partition=batch\n"
            "#SBATCH --time=0-01:00:00\n"
            "#SBATCH --nodes=1\n"
            "#SBATCH --ntasks=1\n"
            "#SBATCH --ntasks-per-node=1\n"
            "#SBATCH --gpus-per-node=1\n"
            "#SBATCH --mem=1\n"
            "#SBATCH --reservation=test-reservation\n"
            "#SBATCH --constraint=gpu\n"
            "#SBATCH --qos=test-qos"
        )
        resources = config.resources.Resources(
            queue="batch",
            node_constraints="gpu",
            runtime="1h",
            nodes=1,
            total_cpus=1,
            cpus_per_node=1,
            gpus_per_node=1,
            memory_per_node="1",
            reservation="test-reservation",
            qos="test-qos",
        )
        result = resources.to_slurm_batch_script()
        assert result == expected

    @pytest.mark.parametrize(
        ("time_str", "expected"),
        [
            ("1s", "0-00:00:01"),
            ("30s", "0-00:00:30"),
            ("60s", "0-00:01:00"),
            ("1m", "0-00:01:00"),
            ("30m", "0-00:30:00"),
            ("60m", "0-01:00:00"),
            ("1h", "0-01:00:00"),
            ("12h", "0-12:00:00"),
            ("24h", "1-00:00:00"),
            ("1d", "1-00:00:00"),
            ("2d", "2-00:00:00"),
        ],
    )
    def test_convert_to_slurm_time_format(self, time_str: str, expected):
        result = config.resources._convert_to_slurm_time_format(time_str)

        assert result == expected

    @pytest.mark.parametrize(
        ("time_str", "expected"),
        [
            (
                "1",
                "Unknown runtime unit '1'. Supported time units: s, m, h, d.",
            ),
            (
                "1h30m",
                (
                    "Invalid runtime value. "
                    "Only single unit allowed (30m, 12h, 1d, etc.)"
                ),
            ),
        ],
    )
    def test_convert_to_slurm_time_format_raises(
        self, expect_raise_if_exception, time_str: str, expected
    ):
        with expect_raise_if_exception(ValueError()) as e:
            config.resources._convert_to_slurm_time_format(time_str)

        assert str(e.value) == expected



import click

import mantik.cli.main as main
import mantik.tracking.track as track


@main.cli.command("init")
@click.option(
    "--no-export",
    is_flag=True,
    show_default=True,
    default=False,
    help="Strip the `export` prefix of the output.",
)
def initialize_tracking(no_export: bool) -> None:
    """Initialize the authentication to mantik and
    print the required environment variables.

    It is not possible to set environment variables in a parent process
    (or shell) from a subprocess (e.g. Python). Thus, the `init` command prints
    the bash export statement with the required environment variable for
    authentication to mantik.
    As a consequence, the output can be directly used to set
    the environment variable in the parent process by using the `eval` bash
    command:

    \b
    ```shell
    eval $(mantik init)
    ```

    If you do not want to set the environment variable in the parent process
    but only want to pass it to the command context, use the `env` bash command
    combined with the `--no-export` flag:

    \b
    ```shell
    env $(mantik init --no-export) <command>
    ```

    """
    environment = track.init_tracking()

    click.echo(environment.to_bash_statement(no_export=no_export))



import pathlib

import mantik.testing as testing
import mantik_compute_backend.firecrest._properties as _properties
import mantik_compute_backend.firecrest.job as _job


class TestJob:
    def test_wait(self):
        job = _job.Job(
            client=testing.firecrest.FakeClient(),
            machine="test-machine",
            job_id="test-job-id",
            job_dir=pathlib.Path("test-rundir"),
        )
        job.wait()
        assert job.status == _properties.Status.COMPLETED



import pytest

import mantik.config._utils as _utils
import mantik.config.exceptions as exceptions


@pytest.mark.parametrize(
    ("name", "value_type", "config", "expected"),
    [
        ("given", str, {"given": "value"}, "value"),
        (
            "given_with_incorrect_type",
            int,
            {"given_with_incorrect_type": "value"},
            exceptions.ConfigValidationError(),
        ),
        (
            "not_given",
            str,
            {"given": "value"},
            exceptions.ConfigValidationError(),
        ),
    ],
)
def test_get_required_config_value(
    expect_raise_if_exception, name, value_type, config, expected
):
    with expect_raise_if_exception(expected):
        result = _utils.get_required_config_value(
            name=name,
            value_type=value_type,
            config=config,
        )

        assert result == expected


@pytest.mark.parametrize(
    ("name", "value_type", "default", "config", "expected"),
    [
        ("given", str, None, {"given": "value"}, "value"),
        (
            "given_with_incorrect_type",
            int,
            None,
            {"given_with_incorrect_type": "value"},
            exceptions.ConfigValidationError(),
        ),
        ("not_given", str, None, {"given": "value"}, None),
        ("not_given", str, None, {"given": "value"}, None),
        ("not_given", str, "default", {"given": "value"}, "default"),
    ],
)
def test_get_optional_config_value(
    expect_raise_if_exception,
    name,
    value_type,
    default,
    config,
    expected,
    caplog,
):
    with expect_raise_if_exception(expected):
        _utils.logger.setLevel("DEBUG")
        result = _utils.get_optional_config_value(
            name=name,
            value_type=value_type,
            config=config,
            default=default,
        )

        assert result == expected
        if expected is None:
            assert "DEBUG" in caplog.text



import mlflow.environment_variables
import pytest

import mantik.mlflow
import mantik.utils as utils


@pytest.fixture(autouse=True)
def set_mlflow_env_vars() -> None:
    with utils.env.env_vars_set(
        {
            utils.mlflow.TRACKING_URI_ENV_VAR: "https://invalid",
            mlflow.environment_variables.MLFLOW_HTTP_REQUEST_TIMEOUT.name: "0",
            mlflow.environment_variables.MLFLOW_HTTP_REQUEST_BACKOFF_FACTOR.name: "0",  # noqa: E501
            mlflow.environment_variables.MLFLOW_HTTP_REQUEST_MAX_RETRIES.name: "0",  # noqa: E501
        }
    ):
        yield


def test_init_tracking(caplog):
    mantik.mlflow.init_tracking()

    assert "Unable to retrieve access token from Mantik API" in caplog.text


def test_start_run(caplog):
    with mantik.mlflow.start_run():
        pass
    assert "Start MLflow run failed" in caplog.text


def test_active_run(caplog):
    assert mantik.mlflow.active_run() is None


def test_end_run(caplog):
    mantik.mlflow.end_run()


def test_delete_run(caplog):
    with pytest.raises(NotImplementedError):
        mantik.mlflow.delete_run("id")


def test_create_experiment(caplog):
    with pytest.raises(NotImplementedError):
        mantik.mlflow.create_experiment("name")


def test_delete_experiment(caplog):
    with pytest.raises(NotImplementedError):
        mantik.mlflow.delete_experiment(0)


def test_log_artifact(caplog, tmp_path):
    mantik.mlflow.log_artifact(tmp_path)
    _assert_expected_error_in_logs(caplog)


def test_log_artifacts(caplog, tmp_path):
    mantik.mlflow.log_artifacts(tmp_path)
    _assert_expected_error_in_logs(caplog)


def test_log_figure(caplog, tmp_path):
    mantik.mlflow.log_figure(tmp_path, "filename")
    _assert_expected_error_in_logs(caplog)


def test_log_image(caplog, tmp_path):
    mantik.mlflow.log_image(tmp_path, "filename")
    _assert_expected_error_in_logs(caplog)


def test_log_metric(caplog):
    mantik.mlflow.log_metric(key="key", value=0, step=0)
    _assert_expected_error_in_logs(caplog)


def test_log_metrics(caplog):
    mantik.mlflow.log_metrics({"key": 0}, step=0)
    _assert_expected_error_in_logs(caplog)


def test_log_param(caplog):
    mantik.mlflow.log_param(key="key", value="value")
    _assert_expected_error_in_logs(caplog)


def test_log_params(caplog):
    mantik.mlflow.log_params({"key": "value"})
    _assert_expected_error_in_logs(caplog)


def test_log_text(caplog):
    mantik.mlflow.log_text("text", "filename")
    _assert_expected_error_in_logs(caplog)


def test_register_model(caplog):
    mantik.mlflow.register_model("uri", "modelname")
    _assert_expected_error_in_logs(caplog)


def test_set_tag(caplog):
    mantik.mlflow.set_tag(key="name", value="value")
    _assert_expected_error_in_logs(caplog)


def test_set_tags(caplog):
    mantik.mlflow.set_tags({"name": "value"})
    _assert_expected_error_in_logs(caplog)


def test_delete_tag(caplog):
    mantik.mlflow.delete_tag("name")
    _assert_expected_error_in_logs(caplog)


def test_set_tracking_uri(caplog):
    mantik.mlflow.set_tracking_uri("mlruns")


def test_get_tracking_uri(caplog):
    result = mantik.mlflow.get_tracking_uri()
    assert result == "mlruns"


@pytest.mark.parametrize(("method"), [mlflow.start_run, "start_run"])
def test_call_method(caplog, method):
    result = mantik.mlflow.call_method(method, "id")

    assert result is None

    assert "Calling mlflow.start_run with args=('id',)" in caplog.text


def _assert_expected_error_in_logs(caplog):
    assert "MLflow API request failed" in caplog.text



import abc
import dataclasses
import logging
import pathlib
import typing as t

import mantik.config._utils as _utils
import mantik.config.exceptions as exceptions
import mantik.utils as utils

logger = logging.getLogger(__name__)

_LOCAL_IMAGE_TYPE = "local"
_REMOTE_IMAGE_TYPE = "remote"
_ALLOWED_IMAGE_TYPES = [_LOCAL_IMAGE_TYPE, _REMOTE_IMAGE_TYPE]


@dataclasses.dataclass
class Execution:
    """Execution environment used for executing a python script."""

    path: str

    @property
    @abc.abstractmethod
    def path_has_to_be_checked(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def get_execution_command(self) -> t.Optional[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_arguments(self) -> t.Optional[t.List[str]]:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, config: t.Dict):
        raise NotImplementedError


@dataclasses.dataclass
class Apptainer(Execution):
    """Information about an Apptainer image.

    Parameters
    ----------
    path : pathlib.Path
        Path to the image.
    type : str, default="local"
        Image type, i.e. if stored locally or remotely.
    options : list[str], optional
        Options to pass to the Apptainer executable.

        .. code-block:: shell
           :caption: Options passed to apptainer run

            srun apptainer run --nv -B /data:/data image.sif

    """

    path: pathlib.Path
    type: str = _LOCAL_IMAGE_TYPE
    options: t.Optional[t.List[str]] = None

    def __post_init__(self) -> None:
        """Check that the given type is correct."""
        self.type = self.type.lower()
        self._ensure_valid_path_type()
        self._ensure_path_conform_to_type()

    @property
    def path_has_to_be_checked(self) -> bool:
        return self.type == _LOCAL_IMAGE_TYPE

    def _ensure_valid_path_type(self) -> None:
        if self.type not in _ALLOWED_IMAGE_TYPES:
            raise exceptions.ConfigValidationError(
                f"The given Apptainer image type "
                f"{self.type!r} is not supported, the supported ones are: "
                f"{utils.formatting.iterable_to_string(_ALLOWED_IMAGE_TYPES)}."  # noqa E501
            )

    def _ensure_path_conform_to_type(self) -> None:
        if self._remote_image_and_relative_path_given():
            raise exceptions.ConfigValidationError(
                f"If image type {self.type!r} is given for an "
                f"Apptainer image, the given path must be "
                f"absolute. The given path is: {self.path.as_posix()!r}."
            )
        elif self._local_image_and_relative_path_given():
            logger.warning(
                (
                    "Path for Apptainer image of type %s is assumed to be "
                    "relative to the MLflow project directory"
                ),
                _LOCAL_IMAGE_TYPE,
            )

    def _remote_image_and_relative_path_given(self) -> bool:
        return self.type == _REMOTE_IMAGE_TYPE and not self.path.is_absolute()

    def _local_image_and_relative_path_given(self) -> bool:
        return self.type == _LOCAL_IMAGE_TYPE and not self.path.is_absolute()

    @classmethod
    def from_dict(cls, config: t.Dict) -> "Apptainer":
        section = _utils.get_required_config_value(
            name="Apptainer",
            value_type=dict,
            config=config,
        )
        path = _utils.get_required_config_value(
            name="Path",
            value_type=pathlib.Path,
            config=section,
        )
        type_ = _utils.get_optional_config_value(
            name="Type",
            value_type=str,
            config=section,
            default=_LOCAL_IMAGE_TYPE,
        )
        options = _utils.get_optional_config_value(
            name="Options",
            value_type=list,
            config=_convert_options_str_to_list(section),
        )
        return cls(
            path=path,
            type=type_,
            options=options,
        )

    @property
    def is_local(self) -> bool:
        """Return whether the image is stored locally."""
        return self.type == _LOCAL_IMAGE_TYPE

    @property
    def is_remote(self) -> bool:
        """Return whether the image is stored remotely."""
        return self.type == _REMOTE_IMAGE_TYPE

    @property
    def name(self) -> str:
        """Return the file name of the image."""
        return self.path.name

    @property
    def path_str(self) -> str:
        """Return the path as a string"""
        return self.path.as_posix()

    def path_as_absolute_to(self, root: pathlib.Path) -> pathlib.Path:
        """Return the image's path as an absolute with the given root."""
        if self.path.is_absolute():
            return self.path
        path = root / self.path
        logger.warning(
            (
                "Assuming that given Apptainer image path %s is relative to "
                "directory %s, hence assuming absolute path %s"
            ),
            self.path,
            root,
            path,
        )
        return path

    def get_execution_command(self) -> str:
        return "srun apptainer"

    def get_arguments(self) -> t.Optional[str]:
        image_path = self._get_image_path()
        options = self.options or []
        return " ".join(["run", *options, image_path])

    def _get_image_path(
        self,
    ) -> str:
        if self.is_local:
            return self.name
        logger.warning(
            "The image is assumed to be already present "
            "in the remote system at the specified path."
        )
        return self.path_str


def _convert_options_str_to_list(config: t.Dict) -> t.Dict:
    # Convert str to list if given for Option
    if "Options" in config and isinstance(config["Options"], str):
        config["Options"] = [config["Options"]]
    return config


@dataclasses.dataclass
class Python(Execution):
    """Python virtual environment used for executing a python script."""

    path: pathlib.Path

    def __post_init__(self) -> None:
        """Check that the given type is correct."""
        self._ensure_venv_path_is_absolute()

    @property
    def path_has_to_be_checked(self) -> bool:
        return False

    @classmethod
    def from_dict(cls, config: t.Dict) -> "Python":
        if isinstance(config["Python"], dict):
            python_dict = _utils.get_required_config_value(
                name="Python",
                value_type=dict,
                config=config,
            )
            python_path = _utils.get_required_config_value(
                name="Path",
                value_type=pathlib.Path,
                config=python_dict,
            )
        else:
            python_path = _utils.get_required_config_value(
                name="Python",
                value_type=pathlib.Path,
                config=config,
            )
        return cls(
            path=python_path,
        )

    def _ensure_venv_path_is_absolute(self) -> None:
        if not self.path.is_absolute():
            raise exceptions.ConfigValidationError(
                "The given path to the Python Venv must be absolute. "
                f"The given path is: {self.path.as_posix()!r}."
            )

    def get_execution_command(self) -> t.Optional[str]:
        return f"source {self.path}/bin/activate"

    def get_arguments(self) -> t.Optional[t.List[str]]:
        return None



import contextlib
import uuid

import pytest
import requests

import mantik.utils.mantik_api.models as models


@pytest.fixture(scope="session")
def model_id() -> uuid.UUID:
    return uuid.uuid4()


@pytest.fixture(scope="session")
def project_id() -> uuid.UUID:
    return uuid.uuid4()


@pytest.fixture(scope="session")
def sample_model_schema(model_id) -> dict:
    return {
        "name": "test-name",
        "modelId": str(model_id),
        "uri": "string",
        "location": "string",
        "connectionId": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
        "mlflowParameters": "{}",
        "status": "PENDING",
        "runId": str(uuid.uuid4()),
    }


@pytest.mark.parametrize(
    ("status_code", "expected"), [(200, None), (404, requests.HTTPError())]
)
def test_get_model(
    mock_mantik_api_request,
    info_caplog,
    status_code,
    expected,
    model_id,
    project_id,
    sample_model_schema,
):
    with mock_mantik_api_request(
        method="GET",
        end_point=f"/projects/{str(project_id)}/"
        f"models/trained/{str(model_id)}",
        status_code=status_code,
        json_response=sample_model_schema,
        expected_error=expected,
    ) as (m, error):
        retrieved_model = models.get_one(
            project_id=project_id, model_id=model_id, token="test_token"
        )
        assert retrieved_model == models.GetModel.from_json(sample_model_schema)
    if error:
        assert any(
            "Call to Mantik API" in message for message in info_caplog.messages
        )


@pytest.mark.parametrize(
    ("status_code", "expected"), [(200, None), (404, requests.HTTPError())]
)
def test_get_all_project_models(
    mock_mantik_api_request,
    info_caplog,
    status_code,
    expected,
    model_id,
    project_id,
    sample_model_schema,
):
    with mock_mantik_api_request(
        method="GET",
        end_point=f"/projects/{str(project_id)}/models/trained",
        status_code=status_code,
        json_response={
            "totalRecords": 1,
            "models": [sample_model_schema],
        },
        expected_error=expected,
    ) as (m, error):
        retrieved_model = models.get_all(
            project_id=project_id, token="test_token"
        )
        assert retrieved_model == [
            models.GetModel.from_json(sample_model_schema)
        ]
    if error:
        assert any(
            "Call to Mantik API" in message for message in info_caplog.messages
        )


@pytest.mark.parametrize(
    ("status_code", "expected"), [(204, None), (404, requests.HTTPError())]
)
def test_delete_model_entry(
    mock_mantik_api_request,
    info_caplog,
    status_code,
    expected,
    model_id,
    project_id,
):
    with mock_mantik_api_request(
        method="DELETE",
        end_point=f"/projects/{str(project_id)}/"
        f"models/trained/{str(model_id)}",
        status_code=status_code,
        json_response={},
        expected_error=expected,
    ) as (m, error):
        models.delete(
            project_id=project_id,
            model_id=model_id,
            token="test_token",
        )
        assert any(
            f"Model with ID: {model_id} has been deleted" in message
            for message in info_caplog.messages
        )
    if error:
        assert any(
            "Call to Mantik API" in message for message in info_caplog.messages
        )


@pytest.mark.parametrize(
    ("model", "error", "extra_logs"),
    [
        (
            models.PostPutModel(
                uri="s3://bucket/item.type",
                location="S3",
                connection_id=uuid.uuid4(),
                mlflow_parameters={},
                name="test-name",
            ),
            None,
            None,
        ),
        (
            models.PostPutModel(
                connection_id=uuid.uuid4(),
                mlflow_parameters={},
                name="test-name",
                run_id=uuid.uuid4(),
            ),
            None,
            None,
        ),
        (
            models.PostPutModel(
                connection_id=uuid.uuid4(),
                mlflow_parameters={},
                name="test-name",
            ),
            models.InvalidModelError,
            None,
        ),
        (
            models.PostPutModel(
                uri="s3://bucket/item.type",
                location="S3",
                connection_id=uuid.uuid4(),
                mlflow_parameters={},
                name="test-name",
                run_id=uuid.uuid4(),
            ),
            None,
            "If both (run ID) and "
            "(uri and location) are defined the (run ID) takes precedence, "
            "and (uri and location) will be overwritten.",
        ),
    ],
)
def test_add_model_entry(
    mock_mantik_api_request,
    info_caplog,
    project_id,
    model_id,
    model,
    error,
    extra_logs,
):
    ctx = contextlib.nullcontext()
    if error:
        ctx = pytest.raises(error)
    with ctx:
        with mock_mantik_api_request(
            method="POST",
            end_point=f"/projects/{str(project_id)}/models/trained",
            status_code=201,
            json_response={"modelId": str(model_id)},
            expected_error=None,
        ):
            new_model_id = models.add(
                new_model_schema=model,
                project_id=project_id,
                token="test_token",
            )
            assert new_model_id == model_id

            assert any(
                f"A model with ID: {model_id} has been created" in message
                for message in info_caplog.messages
            )
            if extra_logs:
                assert extra_logs in info_caplog.messages


@pytest.mark.parametrize(
    ("status_code", "expected"), [(200, None), (404, requests.HTTPError())]
)
def test_update_model(
    mock_mantik_api_request,
    info_caplog,
    status_code,
    expected,
    model_id,
    project_id,
):
    updated_model_schema = models.PostPutModel(
        uri="s3://bucket/item.type",
        location="S3",
        connection_id=uuid.uuid4(),
        mlflow_parameters={},
        run_id=None,
        name="new-name",
    )

    with mock_mantik_api_request(
        method="PUT",
        end_point=f"/projects/{str(project_id)}/"
        f"models/trained/{str(model_id)}",
        status_code=status_code,
        json_response={},
        expected_error=expected,
    ) as (m, error):
        models.update(
            project_id=project_id,
            model_id=model_id,
            updated_model_schema=updated_model_schema,
            token="test_token",
        )
        assert f"Model with ID: {str(model_id)} has been updated"

    if error:
        assert any(
            "Call to Mantik API" in message for message in info_caplog.messages
        )


@pytest.mark.parametrize(
    ("status_code", "expected"), [(200, None), (404, requests.HTTPError())]
)
def test_get_image_url(
    mock_mantik_api_request,
    info_caplog,
    status_code,
    expected,
    model_id,
    project_id,
):
    test_url = "test-url"
    sample_model_schema = {
        "url": test_url,
    }

    with mock_mantik_api_request(
        method="GET",
        end_point=f"/projects/{str(project_id)}/"
        f"models/trained/{str(model_id)}/docker",
        status_code=status_code,
        json_response=sample_model_schema,
        expected_error=expected,
    ) as (m, error):
        url = models.get_image_url(
            project_id=project_id, model_id=model_id, token="test_token"
        )
        assert url == test_url
    if error:
        assert any(
            "Call to Mantik API" in message for message in info_caplog.messages
        )


@pytest.mark.parametrize(
    ("status_code", "expected"), [(200, None), (404, requests.HTTPError())]
)
def test_start_build(
    mock_mantik_api_request,
    info_caplog,
    status_code,
    expected,
    model_id,
    project_id,
):
    with mock_mantik_api_request(
        method="POST",
        end_point=f"/projects/{str(project_id)}/"
        f"models/trained/{str(model_id)}/docker/build",
        status_code=status_code,
        json_response=None,
        expected_error=expected,
    ) as (m, error):
        models.start_build(
            project_id=project_id, model_id=model_id, token="test_token"
        )
    if not error:
        assert any(
            "Building container for model" in message
            for message in info_caplog.messages
        )



import datetime
import uuid

import pytest

import mantik.remote_file_service.data_client as _data_client
import mantik.testing as testing
import mantik.utils.mantik_api.client as mantik_api
import mantik.utils.mantik_api.credentials as _credentials

FAKE_API_URL = "https://fake.com"


@pytest.fixture
def project_id():
    return uuid.UUID("3aeb9774-040e-477e-ae90-878a41c11f80")


@pytest.fixture
def envs():
    _credentials._MANTIK_USERNAME_ENV_VAR = "TEST_MANTIK_USERNAME_ENV_VAR"
    _credentials._MANTIK_PASSWORD_ENV_VAR = "TEST_MANTIK_PASSWORD_ENV_VAR"
    return {
        "TEST_MANTIK_USERNAME_ENV_VAR": "username",
        "TEST_MANTIK_PASSWORD_ENV_VAR": "password",
        "MANTIK_API_USER_ID": "1f981daf-1a2c-4759-a6f1-d25bf8d1d9e7",
        mantik_api._MANTIK_API_URL_ENV_VAR: FAKE_API_URL,
    }


@pytest.fixture
def data_client(
    env_vars_set, envs, requests_mock, mocker, project_id
) -> _data_client.DataClient:
    mocker.patch("pathlib.Path.exists", return_value=False)

    requests_mock.post(
        url=f"{FAKE_API_URL}{mantik_api.MANTIK_API_CREATE_TOKEN_API_PATH}",
        json={
            "AccessToken": "test-access-token",
            "RefreshToken": "test-refresh-token",
            "ExpiresAt": datetime.datetime(2022, 1, 1).isoformat(),
        },
    )

    with env_vars_set(envs):
        yield _data_client.DataClient(
            file_service=testing.remote_file_service.FakeUnicoreFileService(),
        )


@pytest.fixture
def fake_unicore_fs() -> testing.remote_file_service.FakeUnicoreFileService:
    return testing.remote_file_service.FakeUnicoreFileService()



import uuid

import click.testing
import pytest

import mantik.cli.main as main
import mantik.config.core as core
import mantik.testing.token as testing_token
import mantik.utils as utils
import mantik.utils.env_vars as env_vars
import mantik.utils.mantik_api.credentials as _credentials

TEST_MLFLOW_TRACKING_URI = "https://tracking.test-uri.com"
TEST_COMPUTE_BACKEND_URL = "https://compute.test-uri.com"

ENV_VARS = {
    utils.credentials.UNICORE_USERNAME_ENV_VAR: "test-user",
    utils.credentials.UNICORE_PASSWORD_ENV_VAR: "test-password",
    core.COMPUTE_BUDGET_ACCOUNT_ENV_VAR: "test-project",
    utils.mlflow.TRACKING_URI_ENV_VAR: TEST_MLFLOW_TRACKING_URI,
    utils.mlflow.EXPERIMENT_ID_ENV_VAR: "0",
    utils.mlflow.TRACKING_TOKEN_ENV_VAR: "test-token",
    _credentials._MANTIK_USERNAME_ENV_VAR: "mantik-user",
    _credentials._MANTIK_PASSWORD_ENV_VAR: "matik_password",
    env_vars.DATA_REPOSITORY_ID_ENV_VAR: str(uuid.uuid4()),
    env_vars.EXPERIMENT_REPOSITORY_ID_ENV_VAR: str(uuid.uuid4()),
    env_vars.CODE_REPOSITORY_ID_ENV_VAR: str(uuid.uuid4()),
}

SUCCESS_JSON_RESPONSE = {
    "runId": "25d2dad9-dfe2-4971-ad95-b3a9dc363700",
    "timestamp": None,
}


SUCCESS_JSON_RESPONSE_STR = (
    '{"runId": "25d2dad9-dfe2-4971-ad95-b3a9dc363700", "timestamp": null}\n'
)


@testing_token.set_token()
@pytest.mark.parametrize(
    ("cli_args", "expected_exit_code", "expected_output"),
    [
        (
            [
                "--name=test-run",
                "--backend-config=compute-backend-config.json",
                "--branch=main",
                "--entry-point=main",
            ],
            0,
            SUCCESS_JSON_RESPONSE_STR,
        ),
        (
            [
                "--name=test-run",
                "--backend-config=compute-backend-config.json",
                "--branch=main",
                "--entry-point=main",
                "-P a=99",
                "-P b=hello",
            ],
            0,
            SUCCESS_JSON_RESPONSE_STR,
        ),
        ([], 2, "Usage: cli runs submit [OPTIONS] MLPROJECT_PATH\n"),
    ],
)
def test_run_project(
    cli_args,
    expected_exit_code,
    expected_output,
    example_project_absolute_path,
    example_project_relative_path,
    mock_mantik_api_request,
    project_id,
    monkeypatch,
):
    monkeypatch.chdir(example_project_absolute_path)
    with mock_mantik_api_request(
        method="POST",
        end_point=f"/projects/{project_id}/runs",
        status_code=201,
        json_response=SUCCESS_JSON_RESPONSE,
    ) as (m, error):
        with utils.env.env_vars_set(ENV_VARS):
            runner = click.testing.CliRunner()
            result = runner.invoke(
                main.cli,
                [
                    "runs",
                    "submit",
                    example_project_relative_path,
                    *cli_args,
                    f"--project-id={project_id}",
                    f"--connection-id={uuid.uuid4()}",
                ],
            )
        assert result.exit_code == expected_exit_code, result.output
        assert expected_output in result.output


@testing_token.set_token()
def test_run_project_with_absolute_path_for_backend_config(
    example_project_absolute_path,
    example_project_relative_path,
    mock_mantik_api_request,
    project_id,
):
    cli_args = [
        example_project_relative_path,
        f"--backend-config={example_project_absolute_path}/compute-backend-config.json",  # noqa E501
        "--name=test-run",
        "--branch=main",
        f"--connection-id={uuid.uuid4()}",
    ]
    with mock_mantik_api_request(
        method="POST",
        end_point=f"/projects/{project_id}/runs",
        status_code=201,
        json_response=SUCCESS_JSON_RESPONSE,
    ) as (m, error):
        with utils.env.env_vars_set(ENV_VARS):
            runner = click.testing.CliRunner()
            result = runner.invoke(
                main.cli,
                ["runs", "submit", *cli_args, f"--project-id={project_id}"],
            )

        assert result.exit_code == 0, result.output
        assert result.output == SUCCESS_JSON_RESPONSE_STR


@testing_token.set_token()
def test_run_project_with_set_log_level(
    example_project_relative_path,
    example_project_absolute_path,
    caplog,
    mock_mantik_api_request,
    project_id,
):
    with mock_mantik_api_request(
        method="POST",
        end_point=f"/projects/{project_id}/runs",
        status_code=201,
        json_response=SUCCESS_JSON_RESPONSE,
    ) as (m, error):
        with utils.env.env_vars_set(ENV_VARS):
            runner = click.testing.CliRunner()
            _ = runner.invoke(
                main.cli,
                [
                    "runs",
                    "submit",
                    example_project_relative_path,
                    f"--backend-config={example_project_absolute_path}/compute-backend-config.json",  # noqa: E501
                    "--verbose",
                    "--name=test-run",
                    "--branch=main",
                    f"--connection-id={str(uuid.uuid4())}",
                ],
            )
            assert any("DEBUG" in m for m in caplog.messages)



import pathlib

import pytest

import mantik.config

FILE_DIR = pathlib.Path(__file__).parent


@pytest.fixture(scope="function")
def example_unicore_config() -> mantik.config.core.Config:
    return mantik.config.core.Config(
        unicore_api_url="test-url",
        user="user",
        password="password",
        project="test-project",
        environment=mantik.config.environment.Environment(
            execution=mantik.config.executable.Apptainer(
                path=pathlib.Path("mantik-test.sif"),
            )
        ),
        resources=mantik.config.resources.Resources(queue="batch"),
        exclude=["*.sif"],
    )


@pytest.fixture()
def example_project_path() -> pathlib.Path:
    return FILE_DIR / "../../../resources/test-project"



import contextlib
import io
import os
import unittest.mock
import uuid

import mlflow.exceptions
import pytest
import responses

import mantik.runs.local
import mantik.utils.mantik_api.code_repository
import mantik.utils.other


@pytest.fixture()
def mock_get_code(mocker, sample_code_repository):
    mocker.patch(
        "mantik.utils.mantik_api.code_repository.get_one",
        return_value=sample_code_repository,
    )
    yield


@pytest.fixture()
def mock_get_code_from_private_repo(
    mocker, sample_code_repository_with_git_connection
):
    mocker.patch(
        "mantik.utils.mantik_api.code_repository.get_one",
        return_value=sample_code_repository_with_git_connection,
    )
    yield


@pytest.fixture()
def mock_get_experiment(mocker, sample_experiment_repository):
    mocker.patch(
        "mantik.utils.mantik_api.experiment_repository.get_one",
        return_value=sample_experiment_repository,
    )
    yield


@pytest.fixture()
def mock_get_unique_run_name(mocker, sample_run_configuration):
    mocker.patch(
        "mantik.utils.mantik_api.experiment_repository.get_unique_run_name",
        return_value=sample_run_configuration.name,
    )
    yield


@pytest.fixture()
def mock_update_run_status():
    with unittest.mock.patch(
        "mantik.utils.mantik_api.run.update_run_status",
    ) as patch:
        yield patch


@pytest.fixture()
def mock_update_run_logs():
    with unittest.mock.patch(
        "mantik.utils.mantik_api.run.update_logs",
    ) as patch:
        yield patch


@pytest.fixture()
def mock_keyboard_interrupt(mocker):
    mocker.patch("mlflow.run", side_effect=KeyboardInterrupt)
    yield


@pytest.fixture()
def mock_mlflow_execution_exception(mocker):
    mocker.patch(
        "mlflow.run",
        side_effect=mlflow.exceptions.ExecutionException(message="test"),
    )
    yield


@pytest.fixture()
def mock_mlflow_run(mocker):
    class Run:
        def get_status(self):
            return "FINISHED"

    mocker.patch("mlflow.run", return_value=Run())
    yield


@pytest.fixture()
def set_fake_mlflow_tracking():
    os.environ[mantik.utils.mlflow.TRACKING_TOKEN_ENV_VAR] = "FAKE TOKEN"
    os.environ[mantik.utils.mlflow.TRACKING_URI_ENV_VAR] = "mlruns"
    yield
    os.unsetenv(mantik.utils.mlflow.TRACKING_TOKEN_ENV_VAR)
    os.unsetenv(mantik.utils.mlflow.TRACKING_URI_ENV_VAR)


@responses.activate
@pytest.mark.parametrize(
    ("run_output", "expected"),
    [
        (
            mantik.runs.local.LocalRunOutput(
                run_id=uuid.uuid4(), exception=None
            ),
            None,
        ),
        (
            mantik.runs.local.LocalRunOutput(
                run_id=uuid.uuid4(), exception=Exception()
            ),
            Exception,
        ),
    ],
)
def test_local_run(
    mock_get_code,
    mock_get_experiment,
    mock_get_unique_run_name,
    mock_update_run_logs,
    sample_run_configuration,
    mock_get_user_id_from_token,
    mock_code_git_connection_response,
    sample_project_id,
    fake_token,
    run_output,
    expected,
) -> None:
    context = contextlib.nullcontext()
    if expected:
        context = pytest.raises(expected)
    with context:
        mocked_local_run_manager = unittest.mock.Mock(
            spec=mantik.runs.local.LocalRunManager
        )
        mocked_local_run_manager.start_local_run = unittest.mock.Mock(
            return_value=run_output
        )

        mantik.runs.local.run(
            data=sample_run_configuration,
            project_id=sample_project_id,
            mantik_token=fake_token,
            run_manager=mocked_local_run_manager,
            data_target_dir="data",
        )

        mocked_local_run_manager.clone_git_repo.assert_called()
        mocked_local_run_manager.start_local_run.assert_called()


@responses.activate
def test_local_run_with_private_repo(
    mock_get_code_from_private_repo,
    mock_get_experiment,
    mock_get_unique_run_name,
    mock_update_run_logs,
    sample_run_configuration,
    mock_get_user_id_from_token,
    mock_code_git_connection_response,
    sample_project_id,
    fake_token,
) -> None:
    """
    Given I have a code repository in Mantik
    And   it is private
    And   it has a connection attached
    And   that connection has an access token
    When  I execute a local run
    Then  the code repo is cloned
    And   the url is automatically authenticated.
    """
    run_output = mantik.runs.local.LocalRunOutput(
        run_id=uuid.uuid4(), exception=None
    )
    mocked_local_run_manager = unittest.mock.Mock(
        spec=mantik.runs.local.LocalRunManager
    )
    mocked_local_run_manager.start_local_run = unittest.mock.Mock(
        return_value=run_output
    )

    with unittest.mock.patch(
        "mantik.runs.local.construct_git_clone_uri"
    ) as mocked_construct_git_clone_uri:
        mantik.runs.local.run(
            data=sample_run_configuration,
            project_id=sample_project_id,
            mantik_token=fake_token,
            run_manager=mocked_local_run_manager,
            data_target_dir="data",
        )

        mocked_local_run_manager.clone_git_repo.assert_called()
        mocked_local_run_manager.start_local_run.assert_called()
        mocked_construct_git_clone_uri.assert_called()


def test_path_directory_of_mlproject_file() -> None:
    assert (
        mantik.runs.local.path_directory_of_mlproject_file(
            "123/mlproject/MLProject"
        )
        == "123/mlproject"
    )


def test_submit_local_run_and_keyboard_interrupt(
    mock_mantik_api_request,
    sample_run_configuration,
    tmpdir,
    mock_update_run_status,
    mock_keyboard_interrupt,
    set_fake_mlflow_tracking,
) -> None:
    mocked_local_run_manager = mantik.runs.local.LocalRunManager()
    project_id = uuid.uuid4()
    run_id = uuid.uuid4()

    with mock_mantik_api_request(
        method="POST",
        end_point=f"/projects/" f"{str(project_id)}/runs?submit=False",
        status_code=201,
        json_response={"runId": str(run_id)},
    ):
        output = mocked_local_run_manager.start_local_run(
            mlflow_experiment_id=0,
            data=sample_run_configuration,
            mantik_token="access",
            project_id=project_id,
            uri=tmpdir,
        )
        assert isinstance(output.exception, KeyboardInterrupt)
    mlflow.end_run()
    assert mock_update_run_status.call_count == 2
    mock_update_run_status.assert_called_with(
        project_id=project_id, run_id=run_id, token="access", status="KILLED"
    )


def test_submit_local_run_and_execution_exception(
    mock_mantik_api_request,
    sample_run_configuration,
    tmpdir,
    mock_update_run_status,
    mock_mlflow_execution_exception,
    set_fake_mlflow_tracking,
) -> None:
    mocked_local_run_manager = mantik.runs.local.LocalRunManager()
    project_id = uuid.uuid4()
    run_id = uuid.uuid4()
    with mock_mantik_api_request(
        method="POST",
        end_point=f"/projects/" f"{str(project_id)}/runs?submit=False",
        status_code=201,
        json_response={"runId": str(run_id)},
    ):
        output = mocked_local_run_manager.start_local_run(
            mlflow_experiment_id=0,
            data=sample_run_configuration,
            mantik_token="access",
            project_id=project_id,
            uri=tmpdir,
        )
        assert isinstance(
            output.exception, mlflow.exceptions.ExecutionException
        )
    mlflow.end_run()
    assert mock_update_run_status.call_count == 2
    mock_update_run_status.assert_called_with(
        project_id=project_id, run_id=run_id, token="access", status="FAILED"
    )


def test_submit_local_run(
    mock_mantik_api_request,
    sample_run_configuration,
    tmpdir,
    mock_update_run_status,
    mock_mlflow_run,
    set_fake_mlflow_tracking,
) -> None:
    mocked_local_run_manager = mantik.runs.local.LocalRunManager()
    project_id = uuid.uuid4()
    run_id = uuid.uuid4()
    with mock_mantik_api_request(
        method="POST",
        end_point=f"/projects/" f"{str(project_id)}/runs?submit=False",
        status_code=201,
        json_response={"runId": str(run_id)},
    ):
        mocked_local_run_manager.start_local_run(
            mlflow_experiment_id=0,
            data=sample_run_configuration,
            mantik_token="access",
            project_id=project_id,
            uri=tmpdir,
        )
    mlflow.end_run()
    assert mock_update_run_status.call_count == 2
    mock_update_run_status.assert_called_with(
        project_id=project_id, run_id=run_id, token="access", status="FINISHED"
    )


class TestTee:
    def test_write_and_flush(self):
        stream_a = io.StringIO()
        stream_b = io.StringIO()
        tee = mantik.runs.local.Tee(stream_a, stream_b)
        message = "hello to both streams"
        tee.write(message)
        assert stream_a.getvalue() == message
        assert stream_b.getvalue() == message
        tee.flush()
        assert stream_a.read(256) == ""
        assert stream_b.read(256) == ""


@pytest.mark.parametrize(
    ("input_uri", "platform", "git_access_token", "expected_uri"),
    [
        (
            "https://github.com/test/test-private-repository.git",
            "GitHub",
            "ghp_jHtRdboIyHQVqeWpO7LB3irP7WtJlap1rDrI8",
            "https://ghp_jHtRdboIyHQVqeWpO7LB3irP7WtJlap1rDrI8@github.com/test/test-private-repository.git",  # noqa E501
        ),
        (
            "https://gitlab.com/test/test-private-repository.git",
            "GitLab",
            "ghp_jHtRdboIyHQVqeWpO7LB3irP7WtJlap1rDrI8",
            "https://oauth:ghp_jHtRdboIyHQVqeWpO7LB3irP7WtJlap1rDrI8@gitlab.com/test/test-private-repository.git",  # noqa E501
        ),
        (
            "https://gl.test.com/test/test-private-repository.git",
            "GitLab",
            "ghp_jHtRdboIyHQVqeWpO7LB3irP7WtJlap1rDrI8",
            "https://oauth:ghp_jHtRdboIyHQVqeWpO7LB3irP7WtJlap1rDrI8@gl.test.com/test/test-private-repository.git",  # noqa E501
        ),
    ],
)
def test_construct_git_clone_uri(
    input_uri, platform, git_access_token, expected_uri
):
    # test the function with a sample uri
    uri = mantik.utils.other.construct_git_clone_uri(
        uri=input_uri, platform=platform, git_access_token=git_access_token
    )
    assert uri == expected_uri



import dataclasses
import pathlib
import typing as t
import uuid

import mantik.config


@dataclasses.dataclass
class RunConfiguration:
    name: str
    experiment_repository_id: uuid.UUID
    code_repository_id: uuid.UUID
    branch: t.Optional[str]
    commit: t.Optional[str]
    data_repository_id: t.Optional[uuid.UUID]
    mlflow_mlproject_file_path: str
    entry_point: str
    mlflow_parameters: dict
    backend_config: dict
    mlflow_run_id: t.Optional[uuid.UUID] = None
    data_branch: t.Optional[str] = None
    data_commit: t.Optional[str] = None

    def to_post_payload(self) -> dict:
        return {
            "name": self.name,
            "experimentRepositoryId": str(self.experiment_repository_id),
            "codeRepositoryId": str(self.code_repository_id),
            "branch": self.branch,
            "commit": self.commit,
            "dataRepositoryId": str(self.data_repository_id)
            if self.data_repository_id is not None
            else None,
            "dataBranch": self.data_branch if self.data_branch else None,
            "dataCommit": self.data_commit if self.data_commit else None,
            "mlflowMlprojectFilePath": self.mlflow_mlproject_file_path,
            "entryPoint": self.entry_point,
            "mlflowParameters": self.mlflow_parameters,
            "backendConfig": self.backend_config,
            "mlflowRunId": str(self.mlflow_run_id),
        }


@dataclasses.dataclass
class RemoteRunConfiguration:
    name: str
    experiment_repository_id: uuid.UUID
    code_repository_id: uuid.UUID
    branch: t.Optional[str]
    commit: t.Optional[str]
    data_repository_id: t.Optional[uuid.UUID]
    mlflow_mlproject_file_path: str
    entry_point: str
    mlflow_parameters: dict
    backend_config: t.Union[pathlib.Path, str, dict]
    connection_id: t.Optional[uuid.UUID]
    compute_budget_account: t.Optional[str]
    data_branch: t.Optional[str] = None
    data_commit: t.Optional[str] = None

    def to_post_payload(self) -> dict:
        # TODO: the validating has to be fixed
        #  currently it expects all file locally,
        #  probably it just needs to be reduced
        # mantik.config.validate.ProjectValidator(
        #     mlproject_path=self.mlflow_mlproject_file_path,
        #     config=self.backend_config,
        #     mlflow_parameters=self.mlflow_parameters,
        #     entry_point=self.entry_point,
        # ).validate()
        if isinstance(self.backend_config, dict):
            config = self.backend_config
        else:
            config = mantik.config.read.read_config(self.backend_config)

        return {
            "name": self.name,
            "experimentRepositoryId": str(self.experiment_repository_id),
            "codeRepositoryId": str(self.code_repository_id),
            "branch": self.branch,
            "commit": self.commit,
            "dataRepositoryId": str(self.data_repository_id)
            if self.data_repository_id is not None
            else None,
            "dataBranch": self.data_branch if self.data_branch else None,
            "dataCommit": self.data_commit if self.data_commit else None,
            "connectionId": str(self.connection_id),
            "computeBudgetAccount": self.compute_budget_account,
            "mlflowMlprojectFilePath": self.mlflow_mlproject_file_path,
            "entryPoint": self.entry_point,
            "mlflowParameters": self.mlflow_parameters,
            "backendConfig": config,
        }



import logging
import typing as t

import mantik.config.exceptions as exceptions

logger = logging.getLogger(__name__)


def get_required_config_value(
    name: str, value_type: t.Union[t.Type, t.Callable], config: t.Dict
) -> t.Any:
    """Get a required config value.

    Parameters
    ----------
    name : str
        Name of the config value.
    value_type : type
        Python type that the value is expected to have.
        Value will be cast to that type.
    config : t.Dict
        The config to read from.

    Raises
    ------
    ValueError
        If the config value is not present in the config.

    Returns
    -------
    Any
        The value from the config cast as `value_type`.

    """
    if name not in config:
        raise exceptions.ConfigValidationError(
            f"Config is missing entry for key {name!r}"
        )
    return _get_config_value(name=name, value_type=value_type, config=config)


def get_optional_config_value(
    name: str,
    value_type: t.Union[t.Type, t.Callable],
    config: t.Dict,
    default: t.Optional[t.Any] = None,
) -> t.Any:
    """Get an optional config value.

    Parameters
    ----------
    name : str
        Name of the config value.
    value_type : type or callable
        Python type that the value is expected to have.
        Value will be cast to that type.
    config : t.Dict
        The config to read from.
    default : t.Any, default=None
        The default value to return.

    Returns
    -------
    Any or None
        The value from the config cast as `value_type`, `None` if not present.

    """
    if name not in config:
        logger.debug(
            f"'{name}' not in configuration, using default value '{default}'."
        )
        return default
    return _get_config_value(name=name, value_type=value_type, config=config)


def _get_config_value(
    name: str, value_type: t.Union[t.Type, t.Callable], config: t.Dict
) -> t.Any:
    value = config[name]
    return _cast_type(name=name, value=value, value_type=value_type)


def _cast_type(
    name: str, value: str, value_type: t.Union[t.Type, t.Callable]
) -> t.Any:
    try:
        return value_type(value)
    except ValueError as e:
        raise exceptions.ConfigValidationError(
            f"Config value for {name!r} has to be of type {value_type.__name__}"
        ) from e



import dataclasses
import typing as t
import uuid


@dataclasses.dataclass(frozen=True)
class User:
    user_id: uuid.UUID
    name: str

    @classmethod
    def from_json(cls, data: t.Dict) -> "User":
        return cls(
            user_id=uuid.UUID(data["userId"]),
            name=data["name"],
        )
